@@ -9,6 +9,10 @@
 #include "funcno.h"
 #include "kanji.h"
 
+#ifndef	_NOPTY
+#include "termemu.h"
+#endif
+
 #if	MSDOS
 extern int getcurdrv __P_((VOID_A));
 extern int setcurdrv __P_((int, int));
@@ -70,6 +74,10 @@ extern int custno;
 extern int internal_status;
 extern int hideclock;
 extern int fd_restricted;
+#ifndef	_NOPTY
+extern int ptymode;
+extern int parentfd;
+#endif
 
 #ifndef	_NOCOLOR
 static int NEAR getcolorid __P_((namelist *));
@@ -81,16 +89,13 @@ static VOID NEAR statusbar __P_((VOID_A));
 static VOID NEAR stackbar __P_((VOID_A));
 static VOID NEAR cputbytes __P_((off_t, off_t, int));
 static VOID NEAR sizebar __P_((VOID_A));
-#ifndef	NOUID
-static int NEAR putowner __P_((char *, uid_t));
-static int NEAR putgroup __P_((char *, gid_t));
-#endif
 static int NEAR putsize __P_((char *, off_t, int, int));
 static int NEAR putsize2 __P_((char *, namelist *, int));
 static int NEAR putfilename __P_((char *, namelist *, int));
 static VOID NEAR infobar __P_((VOID_A));
 static int NEAR calclocate __P_((int));
 static int NEAR calcfilepos __P_((namelist *, int, char *));
+static int NEAR listupmyself __P_((char *));
 #ifndef	_NOSPLITWIN
 static int NEAR listupwin __P_((char *));
 #endif
@@ -101,7 +106,7 @@ static int readstatus __P_((VOID_A));
 static int NEAR readfilelist __P_((reg_t *, char *));
 static VOID NEAR getfilelist __P_((VOID_A));
 static int NEAR browsedir __P_((char *, char *));
-static char *NEAR initcwd __P_((char *, char *));
+static char *NEAR initcwd __P_((char *, char *, int));
 
 int curcolumns = 0;
 int defcolumns = 0;
@@ -125,13 +130,23 @@ char fullpath[MAXPATHLEN] = "";
 char *macrolist[MAXMACROTABLE];
 int maxmacro = 0;
 int isearch = 0;
-char *helpindex[10] = {
+char *helpindex[MAXHELPINDEX] = {
 #ifdef	_NOTREE
-	"help", "eXec", "Copy", "Delete", "Rename",
-	"Sort", "Find", "Logdir", "Editor", "Unpack"
+	"help",
 #else
-	"Logdir", "eXec", "Copy", "Delete", "Rename",
-	"Sort", "Find", "Tree", "Editor", "Unpack"
+	"Logdir",
+#endif
+	"eXec", "Copy", "Delete", "Rename", "Sort", "Find",
+#ifdef	_NOTREE
+	"Logdir",
+#else
+	"Tree",
+#endif
+	"Editor",
+#ifdef	_NOARCHIVE
+	"",
+#else
+	"Unpack",
 #endif
 };
 char typesymlist[] = "dbclsp";
@@ -164,11 +179,11 @@ int calc_y = -1;
 static CONST u_short modelist[] = {
 	S_IFDIR, S_IFLNK, S_IFSOCK, S_IFIFO, S_IFBLK, S_IFCHR
 };
-#define	MAXMODELIST	((int)sizeof(modelist) / sizeof(u_short))
+#define	MAXMODELIST	arraysize(modelist)
 static CONST char suffixlist[] = {
 	'/', '@', '=', '|'
 };
-#define	MAXSUFFIXLIST	((int)sizeof(suffixlist) / sizeof(char))
+#define	MAXSUFFIXLIST	arraysize(suffixlist)
 #ifndef	_NOCOLOR
 static CONST u_char colorlist[] = {
 	CL_DIR, CL_LINK, CL_SOCK, CL_FIFO, CL_BLOCK, CL_CHAR
@@ -185,7 +200,7 @@ static CONST char defpalette[] = {
 	ANSI_FG,	/* CL_BLOCK */
 	ANSI_FG,	/* CL_CHAR */
 };
-#endif
+#endif	/* !_NOCOLOR */
 #ifndef	_NOPRECEDE
 static int maxstat = 0;
 static int haste = 0;
@@ -246,7 +261,7 @@ static VOID NEAR pathbar(VOID_A)
 		Xputterm(T_STANDOUT);
 		Xcputs2(TS_PATH);
 		Xputterm(END_STANDOUT);
-		Xcprintf2("%-*.*k", TD_PATH, TD_PATH, fullpath);
+		cputstr(TD_PATH, fullpath);
 
 		Xlocate(TC_MARK, TL_PATH);
 		Xcprintf2("%<*d", TD_MARK, mark);
@@ -267,37 +282,42 @@ static VOID NEAR pathbar(VOID_A)
 	Xputterm(T_STANDOUT);
 	Xcputs2(S_PATH);
 	Xputterm(END_STANDOUT);
-	Xcprintf2("%-*.*k", D_PATH, D_PATH, fullpath);
+	cputstr(D_PATH, fullpath);
 
 	Xtflush();
 }
 
 VOID helpbar(VOID_A)
 {
-	int i, j, col, gap, width, len, ofs;
+	int i, j, col, gap, width, len, ofs, max, blk, rest;
+
+	max = MAXHELPINDEX;
+	blk = 5;
 
 	if (ishardomit()) {
-		col = n_column + 1;
+		col = n_column;
 		gap = 0;
 	}
 	else if (iswellomit()) {
-		col = n_column + 1;
+		col = n_column;
 		gap = 1;
 	}
 	else if (isrightomit()) {
-		col = n_column + 1;
+		col = n_column;
 		gap = 2;
 	}
 	else if (isleftshift()) {
-		col = n_column + 1;
-		gap = 3;
-	}
-	else {
 		col = n_column;
 		gap = 3;
 	}
-	width = (col - 4 - gap * 2) / 10 - 1;
-	ofs = (col - (width + 1) * 10 - 2) / 2;
+	else {
+		col = n_column - 1;
+		gap = 3;
+	}
+
+	rest = max - 1 + gap * ((max - 1) / blk);
+	width = (col - 4 - rest) / max;
+	ofs = (n_column - width * max - rest) / 2;
 	if (ofs < 4) ofs = 4;
 
 	Xlocate(0, L_HELP);
@@ -309,13 +329,13 @@ VOID helpbar(VOID_A)
 	Xputch2(isfileflg(dispmode) ? 'F' : ' ');
 #endif
 
-	for (i = 0; i < 10; i++) {
-		Xlocate(ofs + (width + 1) * i + (i / 5) * gap, L_HELP);
-		len = (width - strlen2(helpindex[i])) / 2;
+	for (i = 0; i < max; i++) {
+		Xlocate(ofs + (width + 1) * i + (i / blk) * gap, L_HELP);
+		len = (width - strlen2(helpindex[i]) + 1) / 2;
 		if (len < 0) len = 0;
 		Xputterm(T_STANDOUT);
 		for (j = 0; j < len; j++) Xputch2(' ');
-		Xcprintf2("%-*.*k", width - len, width - len, helpindex[i]);
+		cputstr(width - len, helpindex[i]);
 		Xputterm(END_STANDOUT);
 	}
 
@@ -393,7 +413,7 @@ static VOID NEAR statusbar(VOID_A)
 	Xputterm(T_STANDOUT);
 	Xcputs2(S_FIND);
 	Xputterm(END_STANDOUT);
-	if (findpattern) Xcprintf2("%-*.*k", D_FIND, D_FIND, findpattern);
+	if (findpattern) cputstr(D_FIND, findpattern);
 
 	Xtflush();
 }
@@ -427,11 +447,11 @@ static VOID NEAR stackbar(VOID_A)
 		else
 #endif
 		Xputterm(T_STANDOUT);
-		Xcprintf2("%-*.*k", width - 2, width - 2, filestack[i].name);
+		cputstr(width - 2, filestack[i].name);
 #ifndef	_NOCOLOR
 		x += width - 2;
 		if (bgcolor >= 0) Xchgcolor(bgcolor, 1);
-		else if (ansicolor) Xputterms(T_NORMAL);
+		else if (ansicolor) Xputterm(T_NORMAL);
 		else
 #endif
 		Xputterm(END_STANDOUT);
@@ -466,7 +486,8 @@ static VOID NEAR sizebar(VOID_A)
 	off_t total, fre, bsize;
 
 	if (!hassizeinfo() || !*fullpath) return;
-	if (getinfofs(".", &total, &fre, &bsize) < 0) total = fre = (off_t)-1;
+	if (getinfofs(curpath, &total, &fre, &bsize) < 0)
+		total = fre = (off_t)-1;
 
 #ifndef	_NOTRADLAYOUT
 	if (istradlayout()) {
@@ -539,10 +560,9 @@ int notype;
 
 	i = 0;
 	if (!notype) {
-		for (j = 0; j < (int)sizeof(typelist) / sizeof(u_short); j++)
+		for (j = 0; j < arraysize(typelist); j++)
 			if ((mode & S_IFMT) == typelist[j]) break;
-		buf[i++] = (j < (int)sizeof(typelist) / sizeof(u_short))
-			? typesymlist[j] : '-';
+		buf[i++] = (j < arraysize(typelist)) ? typesymlist[j] : '-';
 	}
 
 	buf[i++] = (mode & S_IRUSR) ? 'r' : '-';
@@ -550,7 +570,7 @@ int notype;
 #if	MSDOS
 	buf[i++] = (mode & S_IXUSR) ? 'x' : '-';
 	buf[i++] = (mode & S_ISVTX) ? 'a' : '-';
-#else
+#else	/* !MSDOS */
 	buf[i++] = (mode & S_ISUID) ? ((mode & S_IXUSR) ? 's' : 'S')
 		: ((mode & S_IXUSR) ? 'x' : '-');
 	buf[i++] = (mode & S_IRGRP) ? 'r' : '-';
@@ -561,7 +581,7 @@ int notype;
 	buf[i++] = (mode & S_IWOTH) ? 'w' : '-';
 	buf[i++] = (mode & S_ISVTX) ? ((mode & S_IXOTH) ? 't' : 'T')
 		: ((mode & S_IXOTH) ? 'x' : '-');
-#endif
+#endif	/* !MSDOS */
 	buf[i] = '\0';
 
 	return(i);
@@ -574,7 +594,7 @@ u_long flags;
 {
 	int i;
 
-	for (i = 0; i < (int)sizeof(fflaglist) / sizeof(u_long); i++)
+	for (i = 0; i < arraysize(fflaglist); i++)
 		buf[i] = (flags & fflaglist[i]) ? fflagsymlist[i] : '-';
 	buf[i] = '\0';
 
@@ -583,7 +603,7 @@ u_long flags;
 #endif
 
 #ifndef	NOUID
-static int NEAR putowner(buf, uid)
+int putowner(buf, uid)
 char *buf;
 uid_t uid;
 {
@@ -598,7 +618,7 @@ uid_t uid;
 	return(len);
 }
 
-static int NEAR putgroup(buf, gid)
+int putgroup(buf, gid)
 char *buf;
 gid_t gid;
 {
@@ -643,7 +663,7 @@ int width;
 		snprintf2(buf, width + 1, "%*.*s", width, width, "<VOL>");
 #else	/* !MSDOS */
 # ifndef	_NODOSDRIVE
-	else if (dospath2("") && s_isfifo(namep))
+	else if (dospath2(nullstr) && s_isfifo(namep))
 		snprintf2(buf, width + 1, "%*.*s", width, width, "<VOL>");
 # endif
 	else if (isdev(namep))
@@ -714,11 +734,14 @@ static VOID NEAR infobar(VOID_A)
 	int len;
 
 	if (!filelist || filepos < 0 || maxfile < 0) return;
+#ifndef	_NOPTY
+	if (parentfd >= 0) return;
+#endif
 
 #ifndef	_NOTRADLAYOUT
 	if (istradlayout()) {
 		Xlocate(TC_INFO + TW_INFO, TL_STATUS);
-		Xcprintf2("%*s", TD_INFO, "");
+		cputspace(TD_INFO);
 
 		Xlocate(TC_INFO + TW_INFO, TL_STATUS);
 		if (filepos >= maxfile) {
@@ -989,7 +1012,7 @@ int no, isstandout;
 			if ((list[no].st_mode & S_IFMT) == modelist[j]) break;
 		if (j < MAXSUFFIXLIST) buf[i] = suffixlist[j];
 		else if (s_isreg(&(list[no]))
-		&& (list[no].st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
+		&& (list[no].st_mode & S_IEXEC_ALL))
 			buf[i] = '*';
 	}
 #ifndef	_NOCOLOR
@@ -1049,7 +1072,7 @@ int no, isstandout;
 		Xchgcolor(bgcolor, 1);
 		Xputch2(' ');
 	}
-	if (ansicolor) Xputterms(T_NORMAL);
+	if (ansicolor) Xputterm(T_NORMAL);
 	else
 #endif
 	if (isstandout > 0) Xputterm(END_STANDOUT);
@@ -1067,7 +1090,7 @@ char *def;
 	if (def) for (i = 0; i < max; i++)
 		if (!strpathcmp(def, list[i].name)) return(i);
 
-	if (list == filelist && max > 1 && !strcmp(list[1].name, ".."))
+	if (list == filelist && max > 1 && isdotdir(list[1].name) == 1)
 		return(1);
 
 	return(0);
@@ -1094,7 +1117,7 @@ int isstandout;
 		if (isstandout) Xputterm(T_STANDOUT);
 		cp = NOFIL_K;
 		if (i <= strlen2(cp)) cp = "NoFiles";
-		Xcprintf2("%-*.*k", i, i, cp);
+		cputstr(i, cp);
 		if (isstandout) Xputterm(END_STANDOUT);
 		win_x = calc_x = i + 2;
 		win_y = calc_y = filetop(win);
@@ -1134,6 +1157,27 @@ int isstandout;
 	return(pos);
 }
 
+static int NEAR listupmyself(def)
+char *def;
+{
+#ifndef	_NOTREE
+	if (treepath) rewritetree();
+	else
+#endif
+#ifndef	_NOCUSTOMIZE
+	if (custno >= 0) rewritecust();
+	else
+#endif
+#ifndef	_NOPTY
+	if (ptylist[win].pid && ptylist[win].status < 0) /*EMPTY*/;
+	else
+#endif
+	if (filelist && (filepos < maxfile || (!filepos && !maxfile)))
+		return(listupfile(filelist, maxfile, def, 1));
+
+	return(-1);
+}
+
 #ifndef	_NOSPLITWIN
 static int NEAR listupwin(def)
 char *def;
@@ -1160,20 +1204,15 @@ char *def;
 	n = -1;
 	for (win = 0; win < windows; win++) {
 		if (win == dupwin) {
-# ifndef	_NOTREE
-			if (treepath) rewritetree();
-			else
-# endif
-# ifndef	_NOCUSTOMIZE
-			if (custno >= 0) rewritecust();
-			else
-# endif
-			if (filelist && filepos < maxfile)
-				n = listupfile(filelist, maxfile, def, 1);
+			n = listupmyself(def);
 			x = win_x;
 			y = win_y;
 		}
-		else if (filelist && filepos < maxfile)
+# ifndef	_NOPTY
+		else if (ptylist[win].pid && ptylist[win].status < 0) continue;
+# endif
+		else if (filelist
+		&& (filepos < maxfile || (!filepos && !maxfile)))
 			listupfile(filelist, maxfile,
 				filelist[filepos].name, 0);
 	}
@@ -1222,6 +1261,9 @@ int n;
 		free(findpattern);
 		findpattern = NULL;
 	}
+# ifndef	_NOPTY
+	killpty(win, NULL);
+# endif
 	win = dupwin;
 
 	return(n);
@@ -1245,19 +1287,19 @@ VOID movepos(old, funcstat)
 int old, funcstat;
 {
 #ifndef	_NOSPLITWIN
-	if ((funcstat & REWRITEMODE) >= REWIN) {
+	if (rewritemode(funcstat) >= FN_REWIN) {
 		filepos = listupwin(filelist[filepos].name);
 		keyflush();
 	}
 	else
 #endif
-	if (((funcstat & REWRITEMODE) >= RELIST)
+	if (rewritemode(funcstat) >= FN_RELIST
 	|| old / FILEPERPAGE != filepos / FILEPERPAGE) {
 		filepos = listupfile(filelist, maxfile,
 			filelist[filepos].name, 1);
 		keyflush();
 	}
-	else if (((funcstat & REWRITEMODE) >= REWRITE) || old != filepos) {
+	else if (rewritemode(funcstat) >= FN_REWRITE || old != filepos) {
 		if (old != filepos) putname(filelist, old, -1);
 		calc_x += putname(filelist, filepos, 1) + 1;
 		win_x = calc_x;
@@ -1269,7 +1311,15 @@ int old, funcstat;
 VOID rewritefile(all)
 int all;
 {
+	int x, y;
+
 	if (!filelist || filepos < 0 || maxfile < 0) return;
+#ifndef	_NOPTY
+	if (parentfd >= 0) return;
+#endif
+
+	x = win_x;
+	y = win_y;
 	if (all > 0) {
 		title();
 		helpbar();
@@ -1285,20 +1335,15 @@ int all;
 	pathbar();
 	if (all >= 0) {
 #ifdef	_NOSPLITWIN
-# ifndef	_NOTREE
-		if (treepath) rewritetree();
-		else
-# endif
-# ifndef	_NOCUSTOMIZE
-		if (custno >= 0) rewritecust();
-		else
-# endif
-		if (filelist && filepos < maxfile)
-			listupfile(filelist, maxfile,
-				filelist[filepos].name, 1);
-#else	/* !_NOSPLITWIN */
+		listupmyself(filelist[filepos].name);
+#else
 		listupwin(filelist[filepos].name);
-#endif	/* !_NOSPLITWIN */
+#endif
+	}
+
+	if (!all) {
+		win_x = x;
+		win_y = y;
 	}
 	Xlocate(win_x, win_y);
 	Xtflush();
@@ -1377,7 +1422,7 @@ char *buf;
 	search_x = len;
 	len = Xkanjiputs(str[2 - s - i]);
 	Xputterm(END_STANDOUT);
-	Xcprintf2("%-*.*k", n_column - len - 1, n_column - len - 1, buf);
+	cputstr(n_column - len - 1, buf);
 	if ((search_x += len) >= n_column) search_x = n_column - 1;
 	search_y = L_HELP;
 	if (i) filepos = pos;
@@ -1441,9 +1486,9 @@ char *arcre;
 	}
 #endif
 
-	if (!(dirp = Xopendir("."))) {
+	if (!(dirp = Xopendir(curpath))) {
 		lostcwd(NULL);
-		if (!(dirp = Xopendir("."))) error(".");
+		if (!(dirp = Xopendir(curpath))) error(curpath);
 	}
 
 	while ((dp = searchdir(dirp, re, arcre))) {
@@ -1505,7 +1550,7 @@ static VOID NEAR getfilelist(VOID_A)
 	else
 #endif
 	{
-		blocksize = getblocksize(".");
+		blocksize = getblocksize(curpath);
 		if (sorttype < 100) sorton = sorttype;
 		if (readfilelist(re, arcre)) {
 #ifndef	_NOPRECEDE
@@ -1542,7 +1587,7 @@ char *file, *def;
 	dupsorton = sorton;
 #endif
 #ifndef	_NOCOLOR
-	if (ansicolor) Xputterms(T_NORMAL);
+	if (ansicolor) Xputterm(T_NORMAL);
 #endif
 
 	mark = 0;
@@ -1554,7 +1599,7 @@ char *file, *def;
 	waitmes();
 
 #ifndef	_NOARCHIVE
-	if (archivefile) def = (*file) ? file : "";
+	if (archivefile) def = (*file) ? file : nullstr;
 #endif
 	getfilelist();
 
@@ -1574,7 +1619,7 @@ char *file, *def;
 	old = filepos = listupwin(def);
 #endif
 	funcstat = 0;
-	no = 1;
+	no = FNC_CANCEL;
 
 	keyflush();
 	buf[0] = '\0';
@@ -1609,7 +1654,7 @@ char *file, *def;
 #ifndef	_NOPRECEDE
 		if (haste && !havestat(&(filelist[filepos]))
 		&& (bindlist[i].d_func < 255
-		|| (funclist[bindlist[i].f_func].status & NEEDSTAT))
+		|| (funclist[bindlist[i].f_func].status & FN_NEEDSTATUS))
 		&& getstatus(&(filelist[filepos])) < 0)
 			no = WARNING_BELL;
 		else
@@ -1621,9 +1666,9 @@ char *file, *def;
 #ifndef	_NOARCHIVE
 		else if (archivefile) continue;
 #endif
-		else funcstat = (KILLSTK | RESCRN | REWIN);
+		else funcstat = (FN_KILLSTACK | FN_RESCREEN | FN_REWIN);
 
-		if (funcstat & KILLSTK) {
+		if (funcstat & FN_KILLSTACK) {
 			if (stackdepth > 0) {
 				chgorder = 0;
 				if (!yesno(KILOK_K)) continue;
@@ -1643,52 +1688,65 @@ char *file, *def;
 					filelist[filepos].name, 1);
 				stackbar();
 			}
-#ifndef	_NOARCHIVE
-			else if (archivefile) /*EMPTY*/;
-#endif
 #ifndef	_NOWRITEFS
+# ifndef	_NOARCHIVE
+			else if (archivefile) /*EMPTY*/;
+# endif
 			else if (chgorder && writefs < 1 && no != WRITE_DIR
 			&& !fd_restricted
-			&& (i = writablefs(".")) > 0 && underhome(NULL) > 0) {
+			&& (i = writablefs(curpath)) > 0
+			&& underhome(NULL) > 0) {
 				chgorder = 0;
 				if (yesno(WRTOK_K)) arrangedir(i);
 			}
-#endif
+#endif	/* !_NOWRITEFS */
 		}
 
 		curfilename = filelist[filepos].name;
-		if (maxfile <= 0 && !(funcstat & NO_FILE)) no = 0;
+		if (maxfile <= 0 && !(funcstat & FN_NOFILE)) no = FNC_NONE;
 #ifndef	_NOARCHIVE
-		else if (archivefile && !(funcstat & ARCH)) no = 0;
+		else if (archivefile && !(funcstat & FN_ARCHIVE))
+			no = FNC_NONE;
 #endif
 		else if (no < FUNCLISTSIZ) {
-			if (!fd_restricted || !(funcstat & RESTRICT))
+			if (!fd_restricted || !(funcstat & FN_RESTRICT))
 				no = (*funclist[no].func)(NULL);
 			else {
 				warning(0, RESTR_K);
-				no = 0;
+				no = FNC_NONE;
 			}
 		}
 		else {
-			no = execusercomm(macrolist[no - FUNCLISTSIZ],
+			no = ptyusercomm(macrolist[no - FUNCLISTSIZ],
 				filelist[filepos].name, 0);
-			no = (no < 0) ? 1 :
-				((internal_status < -1) ? 4 : internal_status);
+			no = evalstatus(no);
+#ifndef	_NOPTY
+			if (ptymode && isearch) no = FNC_NONE;
+#endif
 		}
 
+#ifndef	_NOPTY
+		while (ptylist[win].pid && ptylist[win].status < 0) {
+			VOID_C frontend();
+			no = FNC_UPDATE;
+			funcstat &= ~FN_REWRITEMODE;
+			funcstat |= FN_REWIN;
+		}
+#endif	/* !_NOPTY */
+
 #ifndef	_NOPRECEDE
 		if (sorton) haste = 0;
 #endif
-		if (no < 0 || no >= 4) break;
-		if (no == 1 || no == 3) helpbar();
-		if (no < 2) {
+		if (no < FNC_NONE || no >= FNC_EFFECT) break;
+		if (no == FNC_CANCEL || no == FNC_HELPSPOT) helpbar();
+		if (no < FNC_UPDATE) {
 			funcstat = 0;
 			continue;
 		}
 
-		if ((funcstat & REWRITEMODE) != REWRITE || old != filepos)
+		if (rewritemode(funcstat) != FN_REWRITE || old != filepos)
 			fnameofs = 0;
-		if (funcstat & RESCRN) {
+		if (funcstat & FN_RESCREEN) {
 			title();
 			helpbar();
 			rewritefile(-1);
@@ -1707,12 +1765,12 @@ char *file, *def;
 # endif
 			strcpy(file, filelist[filepos].name);
 		}
-		else if (no > 4) {
+		else if (no == FNC_CHDIR) {
 			char *tmp;
 
 			tmp = (filepos >= 0) ? filelist[filepos].name : NULL;
 			if (!(cp = archchdir(tmp))) {
-				if (!tmp) tmp = "..";
+				if (!tmp) tmp = parentpath;
 				warning(-1, tmp);
 				strcpy(file, tmp);
 			}
@@ -1722,22 +1780,26 @@ char *file, *def;
 				strcpy(file, filelist[filepos].name);
 			}
 		}
-		else if (no == 4) {
+		else if (no == FNC_EFFECT) {
 			if (filepos < 0) *file = '\0';
 			else strcpy(file, filelist[filepos].name);
 		}
-		no = 0;
+		no = FNC_NONE;
 	}
 	else
 #endif	/* !_NOARCHIVE */
-	if (no >= 4) {
-		no -= 4;
-		strcpy(file, (maxfile) ? filelist[filepos].name : ".");
+	if (no >= FNC_EFFECT) {
+		no -= FNC_EFFECT;
+		if (!maxfile && !ischgdir(&(filelist[filepos]))) cp = curpath;
+		else cp = filelist[filepos].name;
+		strcpy(file, cp);
 	}
+
+	i = (maxfile || !ischgdir(&(filelist[0]))) ? maxfile : 1;
 #ifndef	_NOARCHIVE
 	if (!archivefile)
 #endif
-	for (i = 0; i < maxfile; i++) {
+	while (i-- > 0) {
 		free(filelist[i].name);
 		filelist[i].name = NULL;
 	}
@@ -1749,15 +1811,17 @@ char *file, *def;
 	return(no);
 }
 
-static char *NEAR initcwd(path, buf)
+static char *NEAR initcwd(path, buf, evaled)
 char *path, *buf;
+int evaled;
 {
 	char *cp, *file;
 	int i;
 
 	if (!path) return(NULL);
 
-	cp = evalpath(strdup2(path), 0);
+	cp = strdup2(path);
+	if (!evaled) cp = evalpath(cp, 0);
 #if	MSDOS
 	if (_dospath(cp)) {
 		if (setcurdrv(*cp, 0) >= 0 && !Xgetwd(fullpath))
@@ -1779,7 +1843,7 @@ char *path, *buf;
 			i = *file;
 			*file = '\0';
 			if (file == cp) {
-				if (chdir2(_SS_) < 0) error(_SS_);
+				if (chdir2(rootpath) < 0) error(rootpath);
 			}
 			else if (chdir2(cp) < 0) {
 				hideclock = 2;
@@ -1801,8 +1865,9 @@ char *path, *buf;
 	return(file);
 }
 
-VOID main_fd(pathlist)
+VOID main_fd(pathlist, evaled)
 char **pathlist;
+int evaled;
 {
 	char *def, *cwd, file[MAXNAMLEN + 1], prev[MAXNAMLEN + 1];
 	int i, argc, ischgdir;
@@ -1821,6 +1886,12 @@ char **pathlist;
 		winvar[i].v_archivedir = NULL;
 # endif
 #endif	/* !_NOSPLITWIN */
+#ifndef	_NOPTY
+		ptylist[i].pid = (p_id_t)0;
+		ptylist[i].path = NULL;
+		ptylist[i].fd = ptylist[i].pipe = -1;
+		ptylist[i].status = 0;
+#endif
 
 #ifndef	_NOARCHIVE
 		archduplp = NULL;
@@ -1852,7 +1923,7 @@ char **pathlist;
 		if (i >= argc) continue;
 
 		chdir2(cwd);
-		def = initcwd(pathlist[i], prev);
+		def = initcwd(pathlist[i], prev, evaled);
 
 #ifndef	_NOSPLITWIN
 		winvar[i].v_fullpath = strdup2(fullpath);
@@ -1879,17 +1950,17 @@ char **pathlist;
 #endif	/* !_NOSPLITWIN */
 	calcwin();
 
-	strcpy(file, ".");
+	copycurpath(file);
 	_chdir2(fullpath);
 
 	for (;;) {
-		if (!def && !strcmp(file, "..")) {
+		if (!def && isdotdir(file) == 1) {
 			strcpy(prev, getbasename(fullpath));
 			if (*prev) def = prev;
-			else strcpy(file, ".");
+			else copycurpath(file);
 		}
 
-		if (strcmp(file, ".")) {
+		if (isdotdir(file) != 2) {
 #ifdef	_USEDOSEMU
 			char buf[MAXPATHLEN];
 #endif
@@ -1916,7 +1987,7 @@ char **pathlist;
 		if (ischgdir) def = NULL;
 		else {
 			strcpy(prev, file);
-			strcpy(file, ".");
+			copycurpath(file);
 			def = prev;
 		}
 	}
@@ -1927,6 +1998,12 @@ char **pathlist;
 	if (findpattern) free(findpattern);
 #else
 	for (i = 0; i < MAXWINDOWS; i++) shutwin(i);
+	windows = 1;
+	win = 0;
+	calcwin();
+#endif
+#ifndef	_NOPTY
+	killallpty();
 #endif
 	Xlocate(0, n_line - 1);
 	Xtflush();
