@@ -16,6 +16,10 @@
 #include "system.h"
 #endif
 
+#ifndef	_NOPTY
+#include "termemu.h"
+#endif
+
 #define	MD5_BUFSIZ	(128 / 32)
 #define	MD5_BLOCKS	16
 
@@ -62,6 +66,9 @@ extern int inruncom;
 #ifndef	_NOKANJICONV
 extern int unicodebuffer;
 #endif
+#ifndef	_NOPTY
+extern int parentfd;
+#endif
 
 static VOID NEAR builtinerror __P_((char *[], char *, int));
 #ifdef	_NOORIGSHELL
@@ -70,6 +77,9 @@ static VOID NEAR hitkey __P_((int));
 #define	hitkey(n)
 #endif
 static VOID NEAR fputsmeta __P_((char *, FILE *));
+#if	!defined (_NOARCHIVE) && !defined(_NOBROWSE)
+static char **NEAR file2argv __P_((FILE *, char *, int));
+#endif
 #ifndef	_NOARCHIVE
 # if	FD >= 2
 static int NEAR getlaunchopt __P_((int, char *[], char *, launchtable *));
@@ -112,6 +122,9 @@ static int NEAR kconv __P_((int, char *[]));
 static int NEAR getinputstr __P_((int, char *[]));
 static int NEAR getyesno __P_((int, char *[]));
 #endif	/* FD >= 2 */
+#if	!MSDOS && (FD >= 2)
+static int NEAR savetty __P_((int, char *[]));
+#endif
 #ifdef	_NOORIGSHELL
 static int NEAR printenv __P_((int, char *[]));
 static int NEAR setalias __P_((int, char *[]));
@@ -136,7 +149,7 @@ static int NEAR loadsource __P_((int, char *[]));
 #endif
 
 static CONST char *builtinerrstr[] = {
-	"",
+	NULL,
 #define	ER_FEWMANYARG	1
 	"Too few or many arguments",
 #define	ER_OUTOFLIMIT	2
@@ -164,7 +177,7 @@ static CONST char *builtinerrstr[] = {
 #define	ER_NOTDUMBTERM	13
 	"Cannot execute on dumb term",
 };
-#define	BUILTINERRSIZ	((int)(sizeof(builtinerrstr) / sizeof(char *)))
+#define	BUILTINERRSIZ	arraysize(builtinerrstr)
 
 static CONST builtintable builtinlist[] = {
 #ifndef	_NOARCHIVE
@@ -200,6 +213,9 @@ static CONST builtintable builtinlist[] = {
 	{getinputstr,	BL_READLINE},
 	{getyesno,	BL_YESNO},
 #endif	/* FD >= 2 */
+#if	!MSDOS && (FD >= 2)
+	{savetty,	BL_SAVETTY},
+#endif
 #ifdef	_NOORIGSHELL
 # if	FD >= 2
 	{printenv,	BL_SET},
@@ -222,7 +238,7 @@ static CONST builtintable builtinlist[] = {
 # endif
 #endif	/* _NOORIGSHELL */
 };
-#define	BUILTINSIZ	((int)(sizeof(builtinlist) / sizeof(builtintable)))
+#define	BUILTINSIZ	arraysize(builtinlist)
 
 
 static VOID NEAR builtinerror(argv, s, n)
@@ -236,7 +252,8 @@ int n;
 	if (argv && argv[0]) fprintf2(stderr, "%k: ", argv[0]);
 	if (s) fprintf2(stderr, "%k: ", s);
 	fprintf2(stderr, "%s.",
-		(n >= 0) ? builtinerrstr[n] : strerror2(duperrno));
+		(n >= 0 && builtinerrstr[n])
+			? builtinerrstr[n] : strerror2(duperrno));
 	fputnl(stderr);
 }
 
@@ -288,6 +305,87 @@ FILE *fp;
 	}
 }
 
+#if	!defined (_NOARCHIVE) && !defined(_NOBROWSE)
+static char **NEAR file2argv(fp, s, whole)
+FILE *fp;
+char *s;
+int whole;
+{
+	char *cp, *line, **argv;
+	ALLOC_T size;
+	int i, j, pc, argc, escape, quote, pqoute, quoted;
+
+	argc = 1;
+	argv = (char **)malloc2(2 * sizeof(char *));
+	argv[0] = strdup2(s);
+	argv[1] = NULL;
+	j = escape = 0;
+	quote = pqoute = quoted = '\0';
+	cp = c_realloc(NULL, 0, &size);
+	while ((line = fgets2(fp, 0))) {
+		if (!escape && !quote && *line == '#') {
+			free(line);
+			continue;
+		}
+		escape = 0;
+		for (i = 0; line[i]; i++) {
+			cp = c_realloc(cp, j + 2, &size);
+			pc = parsechar(&(line[i]), -1,
+				'\0', EA_EOLMETA, &quote, &pqoute);
+			if (pc == PC_CLQUOTE) quoted = line[i];
+			else if (pc == PC_WORD) {
+				cp[j++] = line[i++];
+				cp[j++] = line[i];
+			}
+			else if (pc == PC_SQUOTE || pc == PC_DQUOTE)
+				cp[j++] = line[i];
+			else if (pc == PC_ESCAPE) {
+				if (!line[++i]) {
+					escape = 1;
+					break;
+				}
+				cp[j++] = line[i];
+			}
+			else if (pc == PC_OPQUOTE) quoted = '\0';
+			else if (pc != PC_NORMAL) /*EMPTY*/;
+			else if (!strchr(IFS_SET, line[i])) cp[j++] = line[i];
+			else if (j || quoted) {
+				quoted = cp[j] = '\0';
+				argv = (char **)realloc2(argv,
+					(argc + 2) * sizeof(char *));
+				argv[argc++] = strdup2(cp);
+				argv[argc] = NULL;
+				j = 0;
+			}
+		}
+
+		if (escape);
+		else if (quote) cp[j++] = '\n';
+		else if (j || quoted) {
+			quoted = cp[j] = '\0';
+			argv = (char **)realloc2(argv,
+				(argc + 2) * sizeof(char *));
+			argv[argc++] = strdup2(cp);
+			argv[argc] = NULL;
+			j = 0;
+		}
+		free(line);
+
+		if (!whole && !escape && !quote) break;
+	}
+
+	if (j) {
+		cp[j] = '\0';
+		argv = (char **)realloc2(argv, (argc + 2) * sizeof(char *));
+		argv[argc++] = strdup2(cp);
+		argv[argc] = NULL;
+	}
+	free(cp);
+
+	return(argv);
+}
+#endif	/* !_NOARCHIVE && !_NOBROWSE */
+
 #ifndef	_NOARCHIVE
 # if	FD >= 2
 static int NEAR getlaunchopt(n, argv, opts, lp)
@@ -318,21 +416,21 @@ launchtable *lp;
 	if (c == 'f') {
 		i = countvar(lp -> format);
 		lp -> format = (char **)realloc2(lp -> format,
-			sizeof(char *) * (i + 2));
+			(i + 2) * sizeof(char *));
 		lp -> format[i++] = strdup2(cp);
 		lp -> format[i] = NULL;
 	}
 	else if (c == 'i') {
 		i = countvar(lp -> lignore);
 		lp -> lignore = (char **)realloc2(lp -> lignore,
-			sizeof(char *) * (i + 2));
+			(i + 2) * sizeof(char *));
 		lp -> lignore[i++] = strdup2(cp);
 		lp -> lignore[i] = NULL;
 	}
 	else if (c == 'e') {
 		i = countvar(lp -> lerror);
 		lp -> lerror = (char **)realloc2(lp -> lerror,
-			sizeof(char *) * (i + 2));
+			(i + 2) * sizeof(char *));
 		lp -> lerror[i++] = strdup2(cp);
 		lp -> lerror[i] = NULL;
 	}
@@ -473,7 +571,7 @@ launchtable *lp;
 	}
 
 	if (i < argc) {
-		lp -> format = (char **)malloc2(sizeof(char *) * 2);
+		lp -> format = (char **)malloc2(2 * sizeof(char *));
 		lp -> format[0] = strdup2(argv[i]);
 		lp -> format[1] = NULL;
 	}
@@ -616,11 +714,17 @@ char *argv[];
 			return(-1);
 		}
 		deletelaunch(n);
+# ifndef	_NOPTY
+		sendparent(TE_DELETELAUNCH, n);
+# endif
 
 		return(0);
 	}
 
 	addlaunch(n, &launch);
+# ifndef	_NOPTY
+	sendparent(TE_ADDLAUNCH, n, &launch);
+# endif
 
 	return(0);
 }
@@ -724,11 +828,17 @@ char *argv[];
 			return(-1);
 		}
 		deletearch(n);
+# ifndef	_NOPTY
+		sendparent(TE_DELETEARCH, n);
+# endif
 
 		return(0);
 	}
 
 	addarch(n, &arch);
+# ifndef	_NOPTY
+	sendparent(TE_ADDARCH, n, &arch);
+# endif
 
 	return(0);
 }
@@ -923,9 +1033,8 @@ static char **NEAR readargv(sargv, dargv)
 char **sargv, **dargv;
 {
 	FILE *fp;
-	char *cp, *line, **argv;
-	ALLOC_T size;
-	int i, j, n, argc, dargc, meta, quote, quoted;
+	char *cp, **argv;
+	int n, dargc;
 
 	dargc = countvar(dargv);
 	for (n = 1; sargv[n]; n++) {
@@ -950,80 +1059,10 @@ char **sargv, **dargv;
 			return(NULL);
 		}
 
-		argc = 1;
-		argv = (char **)malloc2(2 * sizeof(char *));
-		argv[0] = strdup2(sargv[0]);
-		argv[1] = NULL;
-		j = meta = 0;
-		quote = quoted = '\0';
-		cp = c_realloc(NULL, 0, &size);
-		while ((line = fgets2(fp, 0))) {
-			if (!meta && !quote && *line == '#') {
-				free(line);
-				continue;
-			}
-			meta = 0;
-			for (i = 0; line[i]; i++) {
-				cp = c_realloc(cp, j + 2, &size);
-				if (line[i] == quote) {
-					quoted = quote;
-					quote = '\0';
-				}
-				else if (iskanji1(line, i)) {
-					cp[j++] = line[i++];
-					cp[j++] = line[i];
-				}
-#  ifdef	CODEEUC
-				else if (isekana(line, i)) {
-					cp[j++] = line[i++];
-					cp[j++] = line[i];
-				}
-#  endif
-				else if (quote) cp[j++] = line[i];
-				else if (line[i] == PMETA) {
-					if (!line[++i]) {
-						meta = 1;
-						break;
-					}
-					cp[j++] = line[i];
-				}
-				else if (line[i] == '\'' || line[i] == '"') {
-					quoted = '\0';
-					quote = line[i];
-				}
-				else if (!strchr(IFS_SET, line[i]))
-					cp[j++] = line[i];
-				else if (j || quoted) {
-					quoted = cp[j] = '\0';
-					argv = (char **)realloc2(argv,
-						(argc + 2) * sizeof(char *));
-					argv[argc++] = strdup2(cp);
-					argv[argc] = NULL;
-					j = 0;
-				}
-			}
-			if (meta);
-			else if (quote) cp[j++] = '\n';
-			else if (j || quoted) {
-				quoted = cp[j] = '\0';
-				argv = (char **)realloc2(argv,
-					(argc + 2) * sizeof(char *));
-				argv[argc++] = strdup2(cp);
-				argv[argc] = NULL;
-				j = 0;
-			}
-			free(line);
-		}
+		argv = file2argv(fp, sargv[0], 1);
 		if (fp != stdin) Xfclose(fp);
 		else clearerr(fp);
-		if (j) {
-			cp[j] = '\0';
-			argv = (char **)realloc2(argv,
-				(argc + 2) * sizeof(char *));
-			argv[argc++] = strdup2(cp);
-			argv[argc] = NULL;
-		}
-		free(cp);
+
 		dargv = readargv(argv, dargv);
 		freevar(argv);
 		if (!dargv) return(NULL);
@@ -1237,7 +1276,7 @@ bindtable *bindp;
 	}
 	else if (argc > i) {
 		if (argv[i][0] != BINDCOMMENT
-		|| bindp -> key < K_F(1) || bindp -> key > K_F(10)) {
+		|| bindp -> key < K_F(1) || bindp -> key > K_F(MAXHELPINDEX)) {
 			bindp -> key = ER_SYNTAXERR;
 			bindp -> f_func = i;
 			return(-1);
@@ -1276,7 +1315,7 @@ char *func1, *func2, *cp;
 		return(-2);
 	}
 
-	if (bind.key < K_F(1) || bind.key > K_F(10)) {
+	if (bind.key < K_F(1) || bind.key > K_F(MAXHELPINDEX)) {
 		if (cp) free(cp);
 	}
 	else if (cp) {
@@ -1306,9 +1345,10 @@ int n;
 {
 	freemacro(bindlist[n].f_func);
 	freemacro(bindlist[n].d_func);
-	if (bindlist[n].key >= K_F(1) && bindlist[n].key <= K_F(10)) {
+	if (bindlist[n].key >= K_F(1)
+	&& bindlist[n].key <= K_F(MAXHELPINDEX)) {
 		free(helpindex[bindlist[n].key - K_F(1)]);
-		helpindex[bindlist[n].key - K_F(1)] = strdup2("");
+		helpindex[bindlist[n].key - K_F(1)] = strdup2(nullstr);
 	}
 	memmove((char *)&(bindlist[n]), (char *)&(bindlist[n + 1]),
 		(MAXBINDTABLE - n) * sizeof(bindtable));
@@ -1334,6 +1374,9 @@ char *argv[];
 			return(-1);
 		}
 		deletekeybind(no);
+#ifndef	_NOPTY
+		sendparent(TE_DELETEKEYBIND, no);
+#endif
 
 		return(0);
 	}
@@ -1347,6 +1390,9 @@ char *argv[];
 		builtinerror(argv, argv[1 - n], ER_OUTOFLIMIT);
 		return(-1);
 	}
+#ifndef	_NOPTY
+	sendparent(TE_ADDKEYBIND, no, &bind, func1, func2, cp);
+#endif
 
 	return(0);
 }
@@ -1356,7 +1402,8 @@ bindtable *bindp;
 {
 	int n;
 
-	if (bindp -> key < K_F(1) || bindp -> key > K_F(10)) return(NULL);
+	if (bindp -> key < K_F(1) || bindp -> key > K_F(MAXHELPINDEX))
+		return(NULL);
 	n = bindp -> key - K_F(1);
 #ifndef	_NOCUSTOMIZE
 	if (!strcmp(helpindex[n], orighelpindex[n])) return(NULL);
@@ -1570,6 +1617,7 @@ devinfo *devp;
 		}
 		free(sp);
 		free(drvlist);
+		free(devp -> name);
 		return(no);
 	}
 	fdtype[no].offset = 0;
@@ -1609,7 +1657,7 @@ devinfo *devp;
 	}
 
 # ifdef	HDDMOUNT
-	if (!strncmp(argv[3], "HDD", sizeof("HDD") - 1)) {
+	if (!strncmp(argv[3], "HDD", strsize("HDD"))) {
 		if (argc > 5) {
 			devp -> drive = ER_FEWMANYARG;
 			devp -> head = 0;
@@ -1759,6 +1807,9 @@ int isset;
 		}
 # endif
 		deletedrv(i);
+# ifndef	_NOPTY
+		sendparent(TE_DELETEDRV, i);
+# endif
 	}
 	else {
 		if (fdtype[i].name) {
@@ -1782,6 +1833,9 @@ int isset;
 			free(dev.name);
 			return(-1);
 		}
+# ifndef	_NOPTY
+		sendparent(TE_INSERTDRV, i, &dev);
+# endif
 	}
 
 	return(0);
@@ -1956,10 +2010,17 @@ char *argv[];
 	}
 	if (!(key.str)) {
 		setkeyseq(key.code, NULL, 0);
+# ifndef	_NOPTY
+		key.len = (u_char)0;
+		sendparent(TE_SETKEYSEQ, &key);
+# endif
 		return(0);
 	}
 
 	setkeyseq(key.code, key.str, key.len);
+# ifndef	_NOPTY
+	sendparent(TE_SETKEYSEQ, &key);
+# endif
 
 	return(0);
 }
@@ -2071,7 +2132,7 @@ char *argv[];
 		skip = 0;
 		for (i = 1; argv[n][i]; i++) {
 			skip = 0;
-			switch(argv[n][i]) {
+			switch (argv[n][i]) {
 				case 'e':
 					if (argv[n][i + 1]) {
 						editor = &(argv[n][i + 1]);
@@ -2125,7 +2186,8 @@ char *argv[];
 
 		i = n;
 		for (; n < argc; n++) if (!strchr(argv[n], '=')) break;
-		if ((f = parsehist((n < argc) ? argv[n] : "!", NULL)) < 0) {
+		f = parsehist((n < argc) ? argv[n] : "!", NULL, '\0');
+		if (f < 0) {
 			builtinerror(argv, argv[n], ER_EVENTNOFOUND);
 			entryhist(0, tmp, 0);
 			free(tmp);
@@ -2168,12 +2230,12 @@ char *argv[];
 	if (!editor) editor = EDITOR;
 
 	if (list) {
-		f = parsehist((n < argc) ? argv[n] : "-16", NULL);
-		l = parsehist((n + 1 < argc) ? argv[n + 1] : "!", NULL);
+		f = parsehist((n < argc) ? argv[n] : "-16", NULL, '\0');
+		l = parsehist((n + 1 < argc) ? argv[n + 1] : "!", NULL, '\0');
 	}
 	else {
-		f = parsehist((n < argc) ? argv[n] : "!", NULL);
-		l = (n + 1 < argc) ? parsehist(argv[n + 1], NULL) : f;
+		f = parsehist((n < argc) ? argv[n] : "!", NULL, '\0');
+		l = (n + 1 < argc) ? parsehist(argv[n + 1], NULL, '\0') : f;
 	}
 	if (f < 0 || l < 0) {
 		if (f < 0) builtinerror(argv, argv[n], ER_EVENTNOFOUND);
@@ -2207,7 +2269,7 @@ char *argv[];
 		}
 		if (!(fp = Xfdopen(fd, "w"))) {
 			builtinerror(argv, path, -1);
-			Xclose(fd);
+			VOID_C Xclose(fd);
 			rmtmpfile(path);
 			free(tmp);
 			return(-1);
@@ -2646,7 +2708,7 @@ char *argv[];
 	}
 	if (!(wastty = isttyiomode)) ttyiomode(1);
 	duppromptstr = promptstr;
-	promptstr = (argc >= 2) ? argv[1] : "";
+	promptstr = (argc >= 2) ? argv[1] : nullstr;
 	s = inputstr(NULL, 0, -1, NULL, -1);
 	promptstr = duppromptstr;
 	if (!wastty) stdiomode();
@@ -2670,7 +2732,7 @@ char *argv[];
 		builtinerror(argv, NULL, ER_NOTDUMBTERM);
 		return(-1);
 	}
-	if (argc < 2) s = "";
+	if (argc < 2) s = nullstr;
 	else s = argv[1];
 
 	if (!(wastty = isttyiomode)) ttyiomode(1);
@@ -2682,6 +2744,38 @@ char *argv[];
 }
 #endif	/* FD >= 2 */
 
+#if	!MSDOS
+int savestdio(reset)
+int reset;
+{
+	int n;
+
+	if (isttyiomode) return(0);
+
+	n = savettyio(reset);
+# ifndef	_NOPTY
+	sendparent(TE_SAVETTYIO, reset);
+# endif
+	return(n);
+}
+
+# if	FD >= 2
+static int NEAR savetty(argc, argv)
+int argc;
+char *argv[];
+{
+	int n;
+
+	n = 0;
+	if (argc > 1 && argv[1][0] == '-'
+	&& argv[1][1] == 'n' && !(argv[1][2]))
+		n++;
+
+	return(savestdio(n));
+}
+# endif	/* FD >= 2 */
+#endif	/* !MSDOS */
+
 #ifdef	_NOORIGSHELL
 static int NEAR printenv(argc, argv)
 int argc;
@@ -2773,6 +2867,9 @@ char *ident, *comm;
 # ifdef	COMMNOCASE
 	if (ident) for (i = 0; ident[i]; i++) ident[i] = tolower2(ident[i]);
 # endif
+# ifndef	 _NOPTY
+	sendparent(TE_ADDALIAS, ident, comm);
+# endif
 
 	return(0);
 }
@@ -2799,6 +2896,9 @@ char *ident;
 	}
 	if (re) regexp_free(re);
 	if (!n) return(-1);
+# ifndef	 _NOPTY
+	sendparent(TE_DELETEALIAS, ident);
+# endif
 
 	return(0);
 }
@@ -2928,6 +3028,9 @@ char *ident, **comm;
 # ifdef	COMMNOCASE
 	if (ident) for (i = 0; ident[i]; i++) ident[i] = tolower2(ident[i]);
 # endif
+# ifndef	 _NOPTY
+	sendparent(TE_ADDFUNCTION, ident, comm);
+# endif
 
 	return(0);
 }
@@ -2944,6 +3047,9 @@ char *ident;
 	memmove((char *)&(userfunclist[i]),
 		(char *)&(userfunclist[i + 1]),
 		(--maxuserfunc - i) * sizeof(userfunctable));
+# ifndef	 _NOPTY
+	sendparent(TE_DELETEFUNCTION, ident);
+# endif
 
 	return(0);
 }
@@ -3068,7 +3174,7 @@ char *argv[];
 		return(0);
 	}
 
-	var = (char **)malloc2(sizeof(char *) * (n + 1));
+	var = (char **)malloc2((n + 1) * sizeof(char *));
 	for (i = 0; i < n; i++) var[i] = list[i];
 	var[i] = NULL;
 	if (addfunction(strdup2(argv[FUNCNAME]), var) < 0) {
@@ -3183,7 +3289,7 @@ char *argv[];
 		builtinerror(argv, NULL, ER_NOTDUMBTERM);
 		return(-1);
 	}
-	if (fd_restricted && (funclist[n].status & RESTRICT)) {
+	if (fd_restricted && (funclist[n].status & FN_RESTRICT)) {
 		fprintf2(stderr, "%s: %k", argv[0], RESTR_K);
 		fputnl(stderr);
 		return(RET_NOTICE);
@@ -3193,6 +3299,12 @@ char *argv[];
 		fputnl(stderr);
 		return(RET_NOTICE);
 	}
+#ifndef	_NOPTY
+	if (parentfd >= 0) {
+		sendparent(TE_INTERNAL, n, argv[1]);
+		return(RET_SUCCESS);
+	}
+#endif
 	ttyiomode(0);
 	internal_status = (*funclist[n].func)(argv[1]);
 	locate(0, n_line - 1);
@@ -3211,7 +3323,7 @@ int flags;
 
 	n = -1;
 	command = skipspace(command);
-	internal_status = -2;
+	internal_status = FNC_FAIL;
 	if (!(argc = getargs(command, &argv))) {
 		freevar(argv);
 		return(-1);
@@ -3317,7 +3429,7 @@ char ***argvp;
 
 	return(argc);
 }
-#endif
+#endif	/* !_NOCOMPLETE */
 
 #ifdef	DEBUG
 VOID freedefine(VOID_A)
@@ -3346,6 +3458,6 @@ VOID freedefine(VOID_A)
 # ifdef	_USEDOSEMU
 	for (i = 0; fdtype[i].name; i++) free(fdtype[i].name);
 # endif
-	for (i = 0; i < 10; i++) free(helpindex[i]);
+	for (i = 0; i < MAXHELPINDEX; i++) free(helpindex[i]);
 }
 #endif
