@@ -47,8 +47,10 @@ extern int tradlayout;
 extern int sizeinfo;
 #ifndef	_NOCOLOR
 extern int ansicolor;
+# if	FD >= 2
 extern char *ansipalette;
-#endif
+# endif
+#endif	/* !_NOCOLOR */
 #ifndef	_NOEDITMODE
 extern char *editmode;
 #endif
@@ -66,7 +68,7 @@ extern char *promptstr;
 #ifndef	_NOORIGSHELL
 extern char *promptstr2;
 #endif
-#if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
+#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
 extern int unicodebuffer;
 #endif
 #ifndef	_NOKANJIFCONV
@@ -84,10 +86,31 @@ extern char *utf8path;
 extern char *utf8macpath;
 extern char *noconvpath;
 #endif	/* !_NOKANJIFCONV */
+#if	FD >= 2
 extern int tmpumask;
+#endif
 #ifndef	_NOORIGSHELL
 extern int dumbshell;
 #endif
+#ifndef	_NOPTY
+extern int ptymode;
+extern char *ptyterm;
+extern int ptymenukey;
+#endif
+#ifndef	_NOLOGGING
+extern char *logfile;
+extern int logsize;
+# ifndef	NOSYSLOG
+extern int usesyslog;
+# endif
+extern int loglevel;
+# ifndef	NOUID
+extern int rootloglevel;
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+extern int thruargs;
+#endif
 extern int wheader;
 extern char fullpath[];
 extern char *origpath;
@@ -100,6 +123,9 @@ extern int curcolumns;
 extern int subwindow;
 extern int win_x;
 extern int win_y;
+#if	FD >= 2
+extern int lcmdline;
+#endif
 extern int calc_x;
 extern int calc_y;
 extern functable funclist[];
@@ -160,6 +186,9 @@ typedef struct _envtable {
 	u_char type;
 } envtable;
 
+#define	env_str(n)	(&(envlist[n].env[FDESIZ]))
+#define	fdenv_str(n)	(envlist[n].env)
+#define	env_type(n)	(envlist[n].type)
 #ifdef	FORCEDSTDC
 #define	def_str(n)	(envlist[n].def.str)
 #define	def_num(n)	(envlist[n].def.num)
@@ -191,7 +220,9 @@ typedef struct _envtable {
 #define	T_KIN		14
 #define	T_KOUT		15
 #define	T_KNAM		16
-#define	T_OCTAL		17
+#define	T_KTERM		17
+#define	T_OCTAL		18
+#define	T_KEYCODE	19
 
 #ifndef	_NOKANJIFCONV
 typedef struct _pathtable {
@@ -356,7 +387,7 @@ static CONST envtable envlist[] = {
 # if	FD >= 2
 	{"FD_ANSIPALETTE", &ansipalette, DEFVAL(ANSIPALETTE), APAL_E, T_CHARP},
 # endif
-#endif
+#endif	/* !_NOCOLOR */
 #ifndef	_NOEDITMODE
 	{"FD_EDITMODE", &editmode, DEFVAL(EDITMODE), EDMD_E, T_EDIT},
 #endif
@@ -380,11 +411,29 @@ static CONST envtable envlist[] = {
 #endif
 #ifndef	_NOORIGSHELL
 	{"FD_PS2", &promptstr2, DEFVAL(PROMPT2), PRM2_E, T_CHARP},
-#endif
-#ifndef	_NOORIGSHELL
 	{"FD_DUMBSHELL", &dumbshell, DEFVAL(DUMBSHELL), DMSHL_E, T_BOOL},
 #endif
-#if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
+#ifndef	_NOPTY
+	{"FD_PTYMODE", &ptymode, DEFVAL(PTYMODE), PTYMD_E, T_BOOL},
+	{"FD_PTYTERM", &ptyterm, DEFVAL(PTYTERM), PTYTM_E, T_CHARP},
+	{"FD_PTYMENUKEY", &ptymenukey, DEFVAL(PTYMENUKEY), PTYKY_E, T_KEYCODE},
+#endif
+#ifndef	_NOLOGGING
+	{"FD_LOGFILE", &logfile, DEFVAL(LOGFILE), LGFIL_E, T_PATH},
+	{"FD_LOGSIZE", &logsize, DEFVAL(LOGSIZE), LGSIZ_E, T_INT},
+# ifndef	NOSYSLOG
+	{"FD_USESYSLOG", &usesyslog, DEFVAL(USESYSLOG), USYLG_E, T_BOOL},
+# endif
+	{"FD_LOGLEVEL", &loglevel, DEFVAL(LOGLEVEL), LGLVL_E, T_INT},
+# ifndef	NOUID
+	{"FD_ROOTLOGLEVEL", &rootloglevel,
+		DEFVAL(ROOTLOGLEVEL), RLGLV_E, T_INT},
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+	{"FD_THRUARGS", &thruargs, DEFVAL(THRUARGS), THARG_E, T_BOOL},
+#endif
+#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
 	{"FD_UNICODEBUFFER", &unicodebuffer,
 		DEFVAL(UNICODEBUFFER), UNBF_E, T_BOOL},
 #endif
@@ -392,9 +441,16 @@ static CONST envtable envlist[] = {
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
 	{"FD_LANGUAGE", &outputkcode, DEFVAL(NOCNV), LANG_E, T_KOUT},
 #endif
+#ifndef	_NOKANJIFCONV
+	{"FD_DEFKCODE", &defaultkcode, DEFVAL(NOCNV), DFKC_E, T_KNAM},
+#endif
 #ifndef	_NOKANJICONV
 	{"FD_INPUTKCODE", &inputkcode, DEFVAL(NOCNV), IPKC_E, T_KIN},
 #endif
+#if	!defined (_NOKANJICONV) && !defined (_NOPTY)
+	{"FD_PTYINKCODE", &ptyinkcode, DEFVAL(NOCNV), PIKC_E, T_KTERM},
+	{"FD_PTYOUTKCODE", &ptyoutkcode, DEFVAL(NOCNV), POKC_E, T_KTERM},
+#endif
 #ifndef	_NOKANJIFCONV
 	{"FD_FNAMEKCODE", &fnamekcode, DEFVAL(NOCNV), FNKC_E, T_KNAM},
 	{"FD_SJISPATH", &sjispath, DEFVAL(SJISPATH), SJSP_E, T_PATHS},
@@ -413,7 +469,7 @@ static CONST envtable envlist[] = {
 	{"FD_NOCONVPATH", &noconvpath, DEFVAL(NOCONVPATH), NCVP_E, T_PATHS},
 #endif	/* !_NOKANJIFCONV */
 };
-#define	ENVLISTSIZ	((int)(sizeof(envlist) / sizeof(envtable)))
+#define	ENVLISTSIZ	arraysize(envlist)
 
 #ifndef	_NOKANJIFCONV
 static pathtable pathlist[] = {
@@ -422,7 +478,7 @@ static pathtable pathlist[] = {
 	{&progpath, NULL, NOCNV, P_STABLE},
 	{&unitblpath, NULL, NOCNV, P_STABLE},
 };
-#define	PATHLISTSIZ	((int)(sizeof(pathlist) / sizeof(pathtable)))
+#define	PATHLISTSIZ	arraysize(pathlist)
 # ifndef	_NOSPLITWIN
 #  ifndef	_NOARCHIVE
 static pathtable archlist[MAXWINDOWS];
@@ -443,7 +499,7 @@ static CONST devinfo mediadescr[] = {
 	{0xf8, "HDD(PC98)", 'N', 98, 0},
 #  endif
 };
-#define	MEDIADESCRSIZ	((int)(sizeof(mediadescr) / sizeof(devinfo)))
+#define	MEDIADESCRSIZ	arraysize(mediadescr)
 # endif	/* _USEDOSEMU */
 static int cs_item = 0;
 static int cs_max = 0;
@@ -476,7 +532,7 @@ char *s;
 {
 	int n;
 
-	if (!sscanf2(s, "%o%$", &n)) return(-1);
+	if (!sscanf2(s, "%<o%$", &n)) return(-1);
 	n &= 0777;
 
 	return(n);
@@ -489,8 +545,8 @@ int no;
 	char *cp;
 	int n;
 
-	cp = getenv2(envlist[no].env);
-	switch (envlist[no].type) {
+	cp = getenv2(fdenv_str(no));
+	switch (env_type(no)) {
 		case T_BOOL:
 			if (!cp) n = def_num(no);
 			else n = (*cp && atoi2(cp)) ? 1 : 0;
@@ -523,8 +579,8 @@ int no;
 			*((char **)(envlist[no].var)) = cp;
 			break;
 		case T_SORT:
-			if (((n = atoi2(cp)) < 0 || (n & 7) > 5)
-			&& (n < 100 || ((n - 100) & 7) > 5))
+			if ((n = atoi2(cp)) < 0 || (n / 100) > MAXSORTINHERIT
+			|| ((n % 100) & ~15) || ((n % 100) & 7) > MAXSORTTYPE)
 				n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
@@ -587,8 +643,9 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
-			*((int *)(envlist[no].var)) =
-				getlang(cp, envlist[no].type - T_KIN);
+		case T_KTERM:
+			n = (1 << (env_type(no) - T_KIN));
+			*((int *)(envlist[no].var)) = getlang(cp, n);
 			break;
 #endif	/* !_NOKANJICONV || (!_NOENGMES && !NOJPNMES) */
 #if	FD >= 2
@@ -596,6 +653,12 @@ int no;
 			if ((n = atooctal(cp)) < 0) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
+#endif
+#ifndef	_NOPTY
+		case T_KEYCODE:
+			if ((n = getkeycode(cp, 0)) < 0) n = def_num(no);
+			*((int *)(envlist[no].var)) = n;
+			break;
 #endif
 		default:
 			if (!cp) cp = def_str(no);
@@ -617,7 +680,10 @@ int stable;
 
 	if (!stable) /*EMPTY*/;
 	else if (pp -> lang == NOCNV) {
-		pp -> last = NULL;
+		if (!(pp -> flags & P_STABLE)) {
+			if (pp -> last) free(pp -> last);
+			pp -> last = NULL;
+		}
 		stable = 0;
 	}
 	else if (pp -> flags & P_STABLE) /*EMPTY*/;
@@ -650,7 +716,10 @@ pathtable *pp;
 		}
 		if (kanjierrno) pp -> lang = DEFCODE;
 	}
-	pp -> last = strdup2(path);
+	if (!(pp -> flags & P_STABLE)) {
+		if (pp -> last) free(pp -> last);
+		pp -> last = strdup2(path);
+	}
 }
 
 static VOID NEAR savepathlang(VOID_A)
@@ -685,18 +754,37 @@ static VOID NEAR evalpathlang(VOID_A)
 		pathconv(&(fulllist[i]));
 	}
 # endif	/* !_NOSPLITWIN */
+# ifndef	_NOPTY
+	changekcode();
+	changeinkcode();
+	changeoutkcode();
+# endif
 }
 #endif	/* !_NOKANJIFCONV */
 
 static VOID NEAR evalheader(VOID_A)
 {
+#ifndef	_NOEXTRAWIN
+	int i;
+#endif
 	int n;
 
 	n = wheader;
 	wheader = WHEADER;
 	if (n == wheader) return;
 
+#ifdef	_NOEXTRAWIN
 	calcwin();
+#else
+	n -= wheader;
+	for (i = 0; i < windows; i++)
+		if (winvar[i].v_fileperrow + n >= WFILEMIN) break;
+	if (i < windows) winvar[i].v_fileperrow += n;
+#endif
+
+#ifndef	_NOPTY
+	changewsize(wheader, windows);
+#endif
 }
 
 VOID evalenv(VOID_A)
@@ -713,6 +801,9 @@ VOID evalenv(VOID_A)
 #endif
 	evalheader();
 	errno = duperrno;
+#if	!MSDOS && !defined (_NOORIGSHELL)
+	if (autosavetty) savestdio(0);
+#endif
 }
 
 #ifdef	DEBUG
@@ -720,7 +811,13 @@ VOID freeenvpath(VOID_A)
 {
 	int i;
 
-	for (i = 0; i < ENVLISTSIZ; i++) switch (envlist[i].type) {
+# ifndef	_NOKANJIFCONV
+	for (i = 0; i < PATHLISTSIZ; i++) {
+		if (pathlist[i].last) free(pathlist[i].last);
+		pathlist[i].last = NULL;
+	}
+# endif
+	for (i = 0; i < ENVLISTSIZ; i++) switch (env_type(i)) {
 		case T_PATH:
 		case T_PATHS:
 			if (*((char **)(envlist[i].var)))
@@ -731,7 +828,7 @@ VOID freeenvpath(VOID_A)
 			break;
 	}
 }
-#endif
+#endif	/* DEBUG */
 
 #ifndef	_NOCUSTOMIZE
 static int NEAR custputs(s)
@@ -765,7 +862,7 @@ int y, w;
 {
 	Xlocate(0, y);
 	Xputterm(T_STANDOUT);
-	Xcprintf2("%*s", w, "");
+	cputspace(w);
 	Xputterm(END_STANDOUT);
 }
 
@@ -1022,9 +1119,8 @@ char **list;
 		savepathlang();
 #endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			setenv2(envlist[i].env, list[i * 2], 0);
-			setenv2(&(envlist[i].env[FDESIZ]),
-				list[i * 2 + 1], 0);
+			setenv2(fdenv_str(i), list[i * 2], 0);
+			setenv2(env_str(i), list[i * 2 + 1], 0);
 			_evalenv(i);
 		}
 #ifndef	_NOKANJIFCONV
@@ -1035,10 +1131,8 @@ char **list;
 	else {
 		list = (char **)malloc2(ENVLISTSIZ * 2 * sizeof(char *));
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			list[i * 2] = strdup2(getshellvar(envlist[i].env, -1));
-			list[i * 2 + 1] =
-				strdup2(getshellvar(&(envlist[i].env[FDESIZ]),
-					-1));
+			list[i * 2] = strdup2(getshellvar(fdenv_str(i), -1));
+			list[i * 2 + 1] = strdup2(getshellvar(env_str(i), -1));
 		}
 	}
 
@@ -1054,9 +1148,9 @@ static VOID NEAR cleanupenv(VOID_A)
 	savepathlang();
 #endif
 	for (i = 0; i < ENVLISTSIZ; i++) {
-		setenv2(envlist[i].env, NULL, 0);
-		cp = (envlist[i].type == T_CHARP) ? def_str(i) : NULL;
-		setenv2(&(envlist[i].env[FDESIZ]), cp, 0);
+		setenv2(fdenv_str(i), NULL, 0);
+		cp = (env_type(i) == T_CHARP) ? def_str(i) : NULL;
+		setenv2(env_str(i), cp, 0);
 		_evalenv(i);
 	}
 #ifndef	_NOKANJIFCONV
@@ -1113,7 +1207,7 @@ int no;
 	int n, p;
 
 	new = NULL;
-	switch (envlist[no].type) {
+	switch (env_type(no)) {
 		case T_BOOL:
 			str[0] = VBOL0_K;
 			str[1] = VBOL1_K;
@@ -1135,8 +1229,11 @@ int no;
 			str[3] = OSIZE_K;
 			str[4] = ODATE_K;
 			str[5] = OLEN_K;
-			p = (n >= 100) ? 1 : 0;
+			p = n / 100;
+			if (p > MAXSORTINHERIT) p = MAXSORTINHERIT;
 			n %= 100;
+			if ((n & 7) > MAXSORTTYPE)
+				n = ((n & ~7) | MAXSORTTYPE);
 			new = strdup2(&(str[n & 7][3]));
 
 			if (n & 7) {
@@ -1214,6 +1311,7 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
+		case T_KTERM:
 			str[NOCNV] = VNCNV_K;
 			str[ENG] = VENG_K;
 #  ifndef	_NOKANJICONV
@@ -1237,16 +1335,21 @@ int no;
 		case T_OCTAL:
 			cp = ascoctal(*((int *)(envlist[no].var)), buf);
 			break;
+# endif
+# ifndef	_NOPTY
+		case T_KEYCODE:
+			cp = getenv2(fdenv_str(no));
+			break;
 # endif
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			break;
 	}
 
-	if (!getenv2(envlist[no].env))
+	if (!getenv2(fdenv_str(no)))
 		cp = new = strcatalloc(strdup2(cp),
 			(cp && *cp) ? VDEF_K : VUDEF_K);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	n = strlen2(cp);
 	if (n > MAXCUSTVAL - 1) n = MAXCUSTVAL - 1;
 	if (new) free(new);
@@ -1263,8 +1366,8 @@ int no;
 	for (n = 0; n < MAXSELECTSTRS; n++) val[n] = n;
 
 	new = NULL;
-	env = &(envlist[no].env[FDESIZ]);
-	switch (envlist[no].type) {
+	env = env_str(no);
+	switch (env_type(no)) {
 		case T_BOOL:
 			n = (*((int *)(envlist[no].var))) ? 1 : 0;
 			str[0] = VBOL0_K;
@@ -1304,8 +1407,8 @@ int no;
 			break;
 		case T_PATH:
 		case T_PATHS:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
-			new = inputcustenvstr(env, 1, cp, 1);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
+			new = inputcustenvstr(env, 1, cp, HST_PATH);
 			if (new == (char *)-1) return(0);
 			cp = new;
 			break;
@@ -1325,8 +1428,11 @@ int no;
 			val[4] = 5;
 			val[5] = 0;
 			val[6] = -1;
-			p = (n >= 100) ? 100 : 0;
+			p = n / 100;
+			if (p > MAXSORTINHERIT) p = MAXSORTINHERIT;
 			n %= 100;
+			if ((n & 7) > MAXSORTTYPE)
+				n = ((n & ~7) | MAXSORTTYPE);
 			tmp = n & ~7;
 			n &= 7;
 			envcaption(env);
@@ -1347,9 +1453,10 @@ int no;
 			str[0] = VSRT0_K;
 			str[1] = VSRT1_K;
 			val[0] = 0;
-			val[1] = 100;
-			if (noselect(&p, 2, 0, str, val)) return(0);
-			n += p;
+			val[1] = 1;
+			if (noselect(&p, MAXSORTINHERIT + 1, 0, str, val))
+				return(0);
+			n += p * 100;
 			cp = int2str(buf, n);
 			break;
 		case T_DISP:
@@ -1468,6 +1575,7 @@ int no;
 		case T_KIN:
 		case T_KNAM:
 		case T_KOUT:
+		case T_KTERM:
 			tmp = 0;
 			str[tmp] = VNCNV_K;
 			val[tmp++] = NOCNV;
@@ -1486,29 +1594,17 @@ int no;
 			val[tmp++] = JIS8;
 			str[tmp] = "ISO-2022-JP";
 			val[tmp++] = JUNET;
+			str[tmp] = "Hex";
+			val[tmp++] = HEX;
 			str[tmp] = "UTF-8";
 			val[tmp++] = UTF8;
 #  endif	/* !_NOKANJICONV */
 			str[tmp] = VUSET_K;
 			val[tmp++] = -1;
-			p = envlist[no].type - T_KIN;
+			p = (1 << (env_type(no) - T_KIN));
 			for (n = 0; n < tmp; n++) {
-				if (val[n] < 0) continue;
-				switch (kanjiiomode[val[n]]) {
-					case L_INPUT:
-						if (p != L_INPUT) continue;
-						break;
-					case L_OUTPUT:
-						if (p == L_OUTPUT) continue;
-						break;
-					case L_FNAME:
-						if (p == L_FNAME) continue;
-						break;
-					default:
-						continue;
-/*NOTREACHED*/
-						break;
-				}
+				if (val[n] < 0 || (kanjiiomode[val[n]] & p))
+					continue;
 				tmp--;
 				memmove(&(str[n]), &(str[n + 1]),
 					(tmp - n) * sizeof(char *));
@@ -1519,12 +1615,18 @@ int no;
 			if (!tmp) return(0);
 			n = *((int *)(envlist[no].var));
 #  ifndef	_NOKANJICONV
-			if (n != O_JIS7 && n != O_JIS8 && n != O_JUNET
-			&& n != M_UTF8)
-				p = 0;
-			else {
-				p = 1;
-				n--;
+			switch (n) {
+				case O_JIS7:
+				case O_JIS8:
+				case O_JUNET:
+				case CAP:
+				case M_UTF8:
+					p = 1;
+					n--;
+					break;
+				default:
+					p = 0;
+					break;
 			}
 #  endif	/* !_NOKANJICONV */
 			envcaption(env);
@@ -1534,24 +1636,33 @@ int no;
 				break;
 			}
 #  ifndef	_NOKANJICONV
-			if (n >= JIS7 && n <= JUNET) {
-				str[0] = VNJIS_K;
-				str[1] = VOJIS_K;
-				val[0] = 0;
-				val[1] = 1;
-				if (noselect(&p, 2, 64, str, val)) return(0);
-				n += p;
+			switch (n) {
+				case JIS7:
+				case JIS8:
+				case JUNET:
+					str[0] = VNJIS_K;
+					str[1] = VOJIS_K;
+					break;
+				case HEX:
+					str[0] = "HEX";
+					str[1] = "CAP";
+					break;
+				case UTF8:
+					str[0] = VUTF8_K;
+					str[1] = VUTFM_K;
+					break;
+				default:
+					tmp = -1;
+					break;
 			}
-			else if (n == UTF8) {
-				str[0] = VUTF8_K;
-				str[1] = VUTFM_K;
+			if (tmp >= 0) {
 				val[0] = 0;
 				val[1] = 1;
 				if (noselect(&p, 2, 64, str, val)) return(0);
 				n += p;
 			}
 #  endif	/* !_NOKANJICONV */
-			str[NOCNV] = "";
+			str[NOCNV] = nullstr;
 #  if	!defined (_NOENGMES) && !defined (_NOJPNMES)
 			str[ENG] = "C";
 #  endif
@@ -1587,17 +1698,31 @@ int no;
 			cp = ascoctal(n, buf);
 			break;
 # endif	/* FD >= 2 */
+# ifndef	_NOPTY
+		case T_KEYCODE:
+			if (!(cp = getenv2(fdenv_str(no)))) cp = nullstr;
+			new = inputcustenvstr(env, 0, cp, -1);
+			if (new == (char *)-1) return(0);
+			if (new && getkeycode(new, 0) < 0) {
+				warning(0, VALNG_K);
+				free(new);
+				return(0);
+			}
+			cp = new;
+			break;
+# endif	/* !_NOPTY */
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			new = inputcustenvstr(env, 0, cp, -1);
 			if (new == (char *)-1) return(0);
 			cp = new;
 			break;
 	}
 
-	if (getshellvar(envlist[no].env, -1)) setenv2(envlist[no].env, cp, 0);
-	else setenv2(env, cp, 0);
+	if (getshellvar(fdenv_str(no), -1)) n = setenv2(fdenv_str(no), cp, 0);
+	else n = setenv2(env, cp, 0);
 	if (new) free(new);
+	if (n < 0) warning(-1, env);
 #ifndef	_NOKANJIFCONV
 	savepathlang();
 #endif
@@ -1619,11 +1744,11 @@ FILE *fp;
 
 	for (i = n = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
-		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i))))
+		&& (cp = getshellvar(env_str(i), -1))
+		&& (env_type(i) != T_CHARP || strenvcmp(cp, def_str(i))))
 			n++;
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& getshellvar(envlist[i].env, -1))
+		&& getshellvar(fdenv_str(i), -1))
 			n++;
 	}
 	if (!n || !fp) return(n);
@@ -1632,16 +1757,16 @@ FILE *fp;
 	fputs("# shell variables definition\n", fp);
 	for (i = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
-		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i)))) {
+		&& (cp = getshellvar(env_str(i), -1))
+		&& (env_type(i) != T_CHARP || strenvcmp(cp, def_str(i)))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", &(envlist[i].env[FDESIZ]), cp);
+			fprintf2(fp, "%s=%s\n", env_str(i), cp);
 			free(cp);
 		}
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& (cp = getshellvar(envlist[i].env, -1))) {
+		&& (cp = getshellvar(fdenv_str(i), -1))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", envlist[i].env, cp);
+			fprintf2(fp, "%s=%s\n", fdenv_str(i), cp);
 			free(cp);
 		}
 	}
@@ -1660,10 +1785,8 @@ FILE *fp;
 
 	for (n = 0; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
-			ident += FDESIZ;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
+			if (!strnenvcmp(argv[n], fdenv_str(i), len[n])) break;
+			if (!strnenvcmp(argv[n], env_str(i), len[n])) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1676,12 +1799,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 2;
 				break;
@@ -1699,7 +1822,7 @@ FILE *fp;
 		else {
 			if (flaglist) flaglist[i] |= f;
 			if ((cp = getshellvar(ident, -1))
-			&& (envlist[i].type != T_CHARP
+			&& (env_type(i) != T_CHARP
 			|| strenvcmp(cp, def_str(i)))) {
 				if (ns++) fputc(' ', fp);
 				cp = killmeta(cp);
@@ -1741,10 +1864,8 @@ FILE *fp;
 
 	for (n = 1; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strenvcmp(argv[n], ident)) break;
-			ident += FDESIZ;
-			if (!strenvcmp(argv[n], ident)) break;
+			if (!strenvcmp(argv[n], fdenv_str(i))) break;
+			if (!strenvcmp(argv[n], env_str(i))) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1757,12 +1878,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 2;
 				break;
@@ -1781,7 +1902,7 @@ FILE *fp;
 		else {
 			if (flaglist) flaglist[i] |= f;
 			if ((cp = getshellvar(ident, -1))
-			&& (envlist[i].type != T_CHARP
+			&& (env_type(i) != T_CHARP
 			|| strenvcmp(cp, def_str(i)))) {
 				if (ns++) fputc(' ', fp);
 				cp = killmeta(cp);
@@ -1834,7 +1955,7 @@ static VOID NEAR cleanupbind(VOID_A)
 	freestrarray(macrolist, maxmacro);
 	maxmacro = 0;
 	copybind(bindlist, origbindlist);
-	copystrarray(helpindex, orighelpindex, NULL, 10);
+	copystrarray(helpindex, orighelpindex, NULL, MAXHELPINDEX);
 }
 
 static int NEAR dispbind(no)
@@ -1844,7 +1965,7 @@ int no;
 	int len, width;
 
 	if (bindlist[no].key < 0) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (bindlist[no].f_func < FUNCLISTSIZ)
@@ -1857,14 +1978,13 @@ int no;
 	else cp2 = NULL;
 	if (!cp2) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 	}
 	else {
 		width = (MAXCUSTVAL - 1) / 2;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 		putsep();
-		Xcprintf2("%-*.*k",
-			MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width, cp2);
+		cputstr(MAXCUSTVAL - 1 - width, cp2);
 	}
 	len = strlen2(cp1);
 	if (len > --width) len = width;
@@ -1978,7 +2098,7 @@ int no;
 				cp = NULL;
 			else cp = macrolist[bindlist[no].f_func - FUNCLISTSIZ];
 			buf = asprintf3(BNDFC_K, str);
-			cp = inputcuststr(buf, 0, cp, 0);
+			cp = inputcuststr(buf, 0, cp, HST_COM);
 			free(buf);
 			if (!cp);
 			else if (!*cp) {
@@ -2015,7 +2135,7 @@ int no;
 			if (bindlist[no].key < 0 || i < FUNCLISTSIZ) cp = NULL;
 			else cp = macrolist[i - FUNCLISTSIZ];
 			buf = asprintf3(BNDDC_K, str);
-			cp = inputcuststr(buf, 0, cp, 0);
+			cp = inputcuststr(buf, 0, cp, HST_COM);
 			free(buf);
 			if (!cp);
 			else if (!*cp) {
@@ -2031,7 +2151,7 @@ int no;
 		}
 	}
 
-	if (key < K_F(1) || key > K_F(10)) cp = NULL;
+	if (key < K_F(1) || key > K_F(MAXHELPINDEX)) cp = NULL;
 	else cp = inputcuststr(FCOMM_K, 0, helpindex[key - K_F(1)], -1);
 	bind.key = (short)key;
 	bind.f_func = (u_char)n1;
@@ -2164,11 +2284,11 @@ int no;
 
 	key.code = keyseqlist[no];
 	if (getkeyseq(&key) < 0 || !(key.len)) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	cp = encodestr(key.str, key.len);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	len = strlen2(cp);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 	free(cp);
@@ -2419,19 +2539,19 @@ int no;
 	int len, width;
 
 	if (no >= maxlaunch) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (!launchlist[no].format) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 	}
 	else {
 		width = (MAXCUSTVAL - 1 - 6) / 2;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 		putsep();
 		width = MAXCUSTVAL - 1 - 6 - width;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].format[0]);
+		cputstr(width, launchlist[no].format[0]);
 		putsep();
 		Xcprintf2("%-2d", (int)(launchlist[no].topskip));
 		putsep();
@@ -2463,7 +2583,7 @@ launchtable *list;
 	}
 	Xcprintf2("%.*k", len, list -> ext + 1);
 	putsep();
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, list -> comm);
+	cputstr(MAXCUSTVAL, list -> comm);
 	putsep();
 	fillline(yy + y++, n_column);
 	nf = ni = ne = 0;
@@ -2493,8 +2613,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> format[nf]);
+			cputstr(MAXCUSTVAL, list -> format[nf]);
 			putsep();
 			nf++;
 		}
@@ -2512,8 +2631,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lignore[ni]);
+			cputstr(MAXCUSTVAL, list -> lignore[ni]);
 			putsep();
 			ni++;
 		}
@@ -2531,8 +2649,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lerror[ne]);
+			cputstr(MAXCUSTVAL, list -> lerror[ne]);
 			putsep();
 			ne++;
 		}
@@ -2549,7 +2666,7 @@ char *prompt, **var;
 	int i, n, max, val[4];
 
 	max = countvar(var);
-	for (i = 0; i < (int)sizeof(val) / sizeof(int); i++) val[i] = i;
+	for (i = 0; i < arraysize(val); i++) val[i] = i;
 
 	list = (namelist *)malloc2((max + 1) * sizeof(namelist));
 	mes = (char **)malloc2((max + 1) * sizeof(char *));
@@ -2590,8 +2707,7 @@ char *prompt, **var;
 		str[2] = (n < max - 1) ? ARDWN_K : NULL;
 		str[3] = ARDEL_K;
 		i = 0;
-		if (noselect(&i, sizeof(val) / sizeof(int), 0, str, val))
-			continue;
+		if (noselect(&i, arraysize(val), 0, str, val)) continue;
 		if (i == 0) {
 			if (!(tmp = inputcuststr(prompt, 1, cp, -1)))
 				continue;
@@ -2646,17 +2762,17 @@ launchtable *list;
 	str[3] = "-i";
 	str[4] = "-e";
 	str[5] = ARDON_K;
-	for (i = 0; i < (int)sizeof(val) / sizeof(int); i++) val[i] = i;
+	for (i = 0; i < arraysize(val); i++) val[i] = i;
 
 	n = 0;
 	for (;;) {
 		verboselaunch(list);
 		envcaption(ARSEL_K);
-		if (noselect(&n, sizeof(val) / sizeof(int), 0, str, val)) {
+		if (noselect(&n, arraysize(val), 0, str, val)) {
 			if (yesno(ARCAN_K)) return(0);
 			continue;
 		}
-		if (n == (int)sizeof(val) / sizeof(int) - 1) {
+		if (n == arraysize(val) - 1) {
 			if (!(list -> format)) {
 				if (!yesno(ARNOT_K)) continue;
 				freevar(list -> lignore);
@@ -2746,7 +2862,8 @@ int no;
 		list.topskip = launchlist[no].topskip;
 		list.bottomskip = launchlist[no].bottomskip;
 
-		list.comm = inputcuststr(LNCHC_K, 0, launchlist[no].comm, 0);
+		list.comm = inputcuststr(LNCHC_K, 0,
+			launchlist[no].comm, HST_COM);
 		if (!(list.comm)) {
 			freelaunch(&list);
 			return(-1);
@@ -2968,18 +3085,13 @@ int no;
 	int len, width;
 
 	if (no >= maxarchive) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	if (!archivelist[no].p_comm) Xcprintf2("%*s", width, "");
-	else Xcprintf2("%-*.*k", width, width, archivelist[no].p_comm);
+	cputstr(width, archivelist[no].p_comm);
 	putsep();
-	if (!archivelist[no].u_comm)
-		Xcprintf2("%*s", MAXCUSTVAL - 1 - width, "");
-	else Xcprintf2("%-*.*k",
-		MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width,
-		archivelist[no].u_comm);
+	cputstr(MAXCUSTVAL - 1 - width, archivelist[no].u_comm);
 	len = (archivelist[no].p_comm) ? strlen2(archivelist[no].p_comm) : 0;
 	if (len > --width) len = width;
 
@@ -3026,7 +3138,7 @@ int no;
 
 	for (;;) {
 		list.p_comm = inputcuststr(PACKC_K, 0,
-			archivelist[no].p_comm, 0);
+			archivelist[no].p_comm, HST_COM);
 		if (!(list.p_comm)) {
 			free(list.ext);
 			return(0);
@@ -3037,7 +3149,7 @@ int no;
 		}
 
 		list.u_comm = inputcuststr(UPCKC_K, 0,
-			archivelist[no].u_comm, 0);
+			archivelist[no].u_comm, HST_COM);
 		if (!(list.u_comm)) {
 			if (list.p_comm) free(list.p_comm);
 			continue;
@@ -3224,24 +3336,24 @@ int no;
 	int i, len, w1, w2, width;
 
 	if (!fdtype[no].name) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	Xcprintf2("%-*.*k", width, width, fdtype[no].name);
+	cputstr(width, fdtype[no].name);
 	putsep();
 
 	w1 = MAXCUSTVAL - 1 - width;
 #  ifdef	HDDMOUNT
 	if (!fdtype[no].cyl) {
-		char buf[sizeof("HDD98 #offset=") + MAXCOLSCOMMA(3)];
+		char buf[strsize("HDD98 #offset=") + MAXCOLSCOMMA(3) + 1];
 
 		strcpy(buf, "HDD");
 		if (isupper2(fdtype[no].head)) strcat(buf, "98");
 		i = strlen(buf);
-		snprintf2(&(buf[i]), sizeof(buf) - i, " #offset=%'Ld",
+		snprintf2(&(buf[i]), (int)sizeof(buf) - i, " #offset=%'Ld",
 			fdtype[no].offset / fdtype[no].sect);
-		Xcprintf2("%-*.*k", w1, w1, buf);
+		cputstr(w1, buf);
 	}
 	else
 #  endif	/* HDDMOUNT */
@@ -3268,8 +3380,7 @@ int no;
 				break;
 		}
 		w1 -= (w2 + 1) * 3;
-		if (i >= MEDIADESCRSIZ) Xcprintf2("%*s", w1, "");
-		else Xcprintf2("%-*.*k", w1, w1, mediadescr[i].name);
+		cputstr(w1, (i < MEDIADESCRSIZ) ? mediadescr[i].name : NULL);
 	}
 	len = strlen2(fdtype[no].name);
 	if (len > --width) len = width;
@@ -3322,7 +3433,8 @@ int no;
 	for (;;) {
 		if (!(dev.name)) {
 			cp = asprintf3(DRDEV_K, dev.drive);
-			dev.name = inputcuststr(cp, 1, fdtype[no].name, 1);
+			dev.name = inputcuststr(cp, 1,
+				fdtype[no].name, HST_PATH);
 			free(cp);
 			if (!(dev.name)) return(0);
 			else if (!*(dev.name)) {
@@ -3563,7 +3675,7 @@ int no;
 # else
 	mes[4] = COVWR_K;
 # endif
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, mes[no]);
+	cputstr(MAXCUSTVAL, mes[no]);
 	len = strlen2(mes[no]);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 
@@ -3588,7 +3700,7 @@ char *file;
 
 	strcpy(path, file);
 	cp = getbasename(path);
-	len = (int)sizeof(path) - 1 - (cp - path);
+	len = strsize(path) - (cp - path);
 	if (len > MAXTNAMLEN) len = MAXTNAMLEN;
 	genrandname(NULL, 0);
 
@@ -3609,7 +3721,7 @@ char *file;
 	}
 	if (!(fpout = Xfdopen(fd, "w"))) {
 		warning(-1, path);
-		Xclose(fd);
+		VOID_C Xclose(fd);
 		Xunlink(path);
 		if (fpin) Xfclose(fpin);
 		return(-1);
@@ -3831,7 +3943,7 @@ int no;
 				copystrarray(macrolist, tmpmacrolist,
 					&maxmacro, tmpmaxmacro);
 				copystrarray(helpindex, tmphelpindex,
-					NULL, 10);
+					NULL, MAXHELPINDEX);
 				copybind(bindlist, tmpbindlist);
 			}
 # ifndef	_NOKEYMAP
@@ -3865,8 +3977,8 @@ int no;
 # endif
 			break;
 		case 2:
-			if (!(file = inputcuststr(FLOAD_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FLOAD_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			if (!*file) {
 				free(file);
 				break;
@@ -3885,8 +3997,8 @@ int no;
 			}
 			break;
 		case 3:
-			if (!(file = inputcuststr(FSAVE_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FSAVE_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			done = 1;
 			file = evalpath(file, 0);
 # ifdef	FAKEUNINIT
@@ -3924,8 +4036,8 @@ int no;
 			break;
 		case 4:
 # ifndef	_NOORIGSHELL
-			if (!(file = inputcuststr(FOVWR_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FOVWR_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			done = 1;
 			file = evalpath(file, 0);
 			if (!*file
@@ -3961,7 +4073,7 @@ int no, y, isstandout;
 	cp = NEWET_K;
 	switch (custno) {
 		case 0:
-			cp = &(envlist[no].env[FDESIZ]);
+			cp = env_str(no);
 			break;
 		case 1:
 			if (bindlist[no].key >= 0)
@@ -4017,7 +4129,7 @@ int no, y, isstandout;
 			cp = name[no];
 			break;
 		default:
-			cp = "";
+			cp = nullstr;
 			break;
 	}
 
@@ -4080,8 +4192,7 @@ static VOID NEAR dispcust(VOID_A)
 					Xcprintf2("%.*k", MAXCUSTVAL, NIMPL_K);
 				if (cs_len[i - start] > MAXCUSTVAL - 1)
 					cs_len[i - start] = MAXCUSTVAL - 1;
-				else Xcprintf2("%*s",
-					MAXCUSTVAL - cs_len[i - start], "");
+				else cputspace(MAXCUSTVAL - cs_len[i - start]);
 				break;
 		}
 		putsep();
@@ -4194,7 +4305,7 @@ int customize(VOID_A)
 
 	tmpenvlist = copyenv(NULL);
 	tmpmacrolist = copystrarray(NULL, macrolist, &tmpmaxmacro, maxmacro);
-	tmphelpindex = copystrarray(NULL, helpindex, NULL, 10);
+	tmphelpindex = copystrarray(NULL, helpindex, NULL, MAXHELPINDEX);
 	tmpbindlist = copybind(NULL, bindlist);
 	for (i = 0; bindlist[i].key >= 0; i++);
 # ifndef	_NOKEYMAP
@@ -4345,7 +4456,7 @@ int customize(VOID_A)
 		free(tmpmacrolist);
 	}
 	if (tmphelpindex) {
-		freestrarray(tmphelpindex, 10);
+		freestrarray(tmphelpindex, MAXHELPINDEX);
 		free(tmphelpindex);
 	}
 	if (tmpbindlist) free(tmpbindlist);
