@@ -65,7 +65,7 @@ extern int errno;
 #define	getconstvar(s)		(char *)getenv(s)
 #define	getshellvar(s, l)	(char *)getenv(s)
 #else
-#define	getconstvar(s)		(getshellvar(s, sizeof(s) - 1))
+#define	getconstvar(s)		(getshellvar(s, strsize(s)))
 #endif
 
 #ifdef	USESTRERROR
@@ -77,8 +77,22 @@ extern char *sys_errlist[];
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
 
+#ifdef	_NOPTY
+#define	isptymode()		0
+#define	isshptymode()		0
+#else	/* !_NOPTY */
+# ifdef	_NOORIGSHELL
+# define	isptymode()	ptymode
+# define	isshptymode()	0
+# else
+# define	isptymode()	(ptymode && (fdmode || !shptymode))
+# define	isshptymode()	(!fdmode && shptymode)
+# endif
+#endif	/* !_NOPTY */
+
 /* main.c */
 extern VOID error __P_((char *));
+extern VOID setlinecol __P_((VOID_A));
 extern VOID checkscreen __P_((int, int));
 #ifdef	SIGWINCH
 extern VOID pollscreen __P_((int));
@@ -89,6 +103,8 @@ extern VOID title __P_((VOID_A));
 extern VOID saveorigenviron __P_((VOID_A));
 #endif
 extern int loadruncom __P_((char *, int));
+extern VOID initfd __P_((char **));
+extern VOID prepareexitfd __P_((int));
 
 /* system.c */
 #ifdef	_NOORIGSHELL
@@ -101,6 +117,23 @@ extern FILE *dopopen __P_((char *));
 extern int dopclose __P_((FILE *));
 #endif
 
+/* log.c */
+#ifdef	_NOLOGGING
+#define	LOG0(l, n, f)
+#define	LOG1(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3)
+#else	/* !_NOLOGGING */
+#define	LOG0(l, n, f)		logsyscall(l, n, f)
+#define	LOG1(l, n, f, a1)	logsyscall(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)	logsyscall(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3) \
+				logsyscall(l, n, f, a1, a2, a3)
+extern VOID logclose __P_((VOID_A));
+extern VOID logsyscall __P_((int, int, CONST char *, ...));
+extern VOID logmessage __P_((int, CONST char *, ...));
+#endif	/* !_NOLOGGING */
+
 /* dosemu.c or unixemu.c */
 #ifdef	_USEDOSPATH
 extern int _dospath __P_((char *));
@@ -122,7 +155,7 @@ extern VOID Xrewinddir __P_((DIR *));
 #if	MSDOS
 extern int rawchdir __P_((char *));
 #else
-#define	rawchdir	chdir
+#define	rawchdir(p)	((chdir(p)) ? -1 : 0)
 #endif
 extern int Xchdir __P_((char *));
 extern char *Xgetwd __P_((char *));
@@ -137,11 +170,17 @@ extern int Xutime __P_((char *, struct utimbuf *));
 #else
 extern int Xutimes __P_((char *, struct timeval []));
 #endif
+#ifdef	HAVEFLAGS
+extern int Xchflags __P_((char *, u_long));
+#endif
+#ifndef	NOUID
+extern int Xchown __P_((char *, uid_t, gid_t));
+#endif
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
 extern int Xopen __P_((char *, int, int));
 #ifdef	_NODOSDRIVE
-#define	Xclose		close
+#define	Xclose(f)	((close(f)) ? -1 : 0)
 #define	Xread		read
 #define	Xwrite		write
 #define	Xlseek		lseek
@@ -164,6 +203,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #ifdef	_NODOSDRIVE
 #define	Xfdopen		fdopen
 #define	Xfclose		fclose
+#define	Xfileno		fileno
 #define	Xfeof		feof
 #define	Xfread		fread
 #define	Xfwrite		fwrite
@@ -175,6 +215,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #else
 extern FILE *Xfdopen __P_((int, char *));
 extern int Xfclose __P_((FILE *));
+extern int Xfileno __P_((FILE *));
 extern int Xfeof __P_((FILE *));
 extern int Xfread __P_((char *, int, int, FILE *));
 extern int Xfwrite __P_((char *, int, int, FILE *));
@@ -194,6 +235,44 @@ extern int Xpclose __P_((FILE *));
 # endif
 #endif	/* _NOORIGSHELL */
 
+/* pty.c */
+#ifndef	_NOPTY
+extern int Xopenpty __P_((int *, char *, ALLOC_T));
+extern int Xlogin_tty __P_((char *, char *, char *));
+extern p_id_t Xforkpty __P_((int *, char *, char *));
+#endif
+
+/* termemu.c */
+#ifdef	_NOPTY
+#define	ptymacro		execmacro
+#define	ptyusercomm		execusercomm
+#define	ptysystem		dosystem
+#else	/* !_NOPTY */
+extern VOID regionscroll __P_((int, int, int, int, int, int));
+extern int selectpty __P_((int, int [], char [], int));
+extern VOID syncptyout __P_((int, int));
+extern int recvbuf __P_((int, VOID_P, int));
+extern VOID sendbuf __P_((int, VOID_P, int));
+extern int recvword __P_((int, int *));
+extern VOID sendword __P_((int, int));
+extern int recvstring __P_((int, char **));
+extern VOID sendstring __P_((int, char *));
+extern VOID sendparent __P_((int, ...));
+extern int ptymacro __P_((char *, char *, int));
+# ifdef	_NOORIGSHELL
+# define	ptyusercomm(c,a,f)	ptymacro(c, a, f | F_EVALMACRO)
+# else
+# define	ptyusercomm		ptymacro
+# endif
+#define	ptysystem(c)		ptymacro(c, NULL, F_DOSYSTEM)
+extern VOID killpty __P_((int, int *));
+extern VOID killallpty __P_((VOID_A));
+extern int checkpty __P_((int));
+extern int checkallpty __P_((VOID_A));
+#endif	/* !_NOPTY */
+
+/* frontend.c */
+#ifdef	_NOPTY
 #define	Xttyiomode		ttyiomode
 #define	Xstdiomode		stdiomode
 #define	Xtermmode		termmode
@@ -209,6 +288,40 @@ extern int Xpclose __P_((FILE *));
 #define	Xkanjiputs		kanjiputs
 #define	Xchgcolor		chgcolor
 #define	Xmovecursor		movecursor
+#else	/* !_NOPTY */
+extern int waitstatus __P_((p_id_t, int, int *));
+extern VOID Xttyiomode __P_((int));
+extern VOID Xstdiomode __P_((VOID_A));
+extern int Xtermmode __P_((int));
+extern VOID Xputch2 __P_((int));
+extern VOID Xcputs2 __P_((char *));
+extern VOID Xputterm __P_((int));
+extern VOID Xputterms __P_((int));
+extern int Xsetscroll __P_((int, int));
+extern VOID Xlocate __P_((int, int));
+extern VOID Xtflush __P_((VOID_A));
+extern int Xcprintf2 __P_((CONST char *, ...));
+extern VOID Xcputnl __P_((VOID_A));
+extern int Xkanjiputs __P_((char *));
+extern VOID Xchgcolor __P_((int, int));
+extern VOID Xmovecursor __P_((int, int, int));
+extern VOID changewin __P_((int, p_id_t));
+extern VOID changewsize __P_((int, int));
+extern VOID insertwin __P_((int, int));
+extern VOID deletewin __P_((int, int));
+# ifndef	_NOKANJICONV
+extern VOID changekcode __P_((VOID_A));
+extern VOID changeinkcode __P_((VOID_A));
+extern VOID changeoutkcode __P_((VOID_A));
+# endif
+extern int frontend __P_((VOID_A));
+#endif	/* !_NOPTY */
+
+/* backend.c */
+#ifndef	_NOPTY
+extern VOID resetptyterm __P_((int, int));
+extern int backend __P_((VOID_A));
+#endif
 
 /* libc.c */
 extern int stat2 __P_((char *, struct stat *));
@@ -267,6 +380,9 @@ extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
 extern struct dirent *searchdir __P_((DIR *, reg_t *, char *));
 extern int underhome __P_((char *));
 extern int preparedir __P_((char *));
+#ifndef	NOFLOCK
+extern int lockfile __P_((int, int));
+#endif
 extern int touchfile __P_((char *, struct stat *));
 #ifdef	_NODOSDRIVE
 #define	nodoslstat	Xlstat
@@ -384,6 +500,9 @@ extern int parsesetdrv __P_((int, char *[], devinfo *));
 extern int parsekeymap __P_((int, char *[], keyseq_t *));
 extern VOID printkeymap __P_((keyseq_t *, int, FILE *));
 #endif
+#if	!MSDOS
+extern int savestdio __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 extern int searchalias __P_((char *, int));
 extern int addalias __P_((char *, char *));
@@ -431,6 +550,7 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_KCONV	"kconv"
 #define	BL_READLINE	"readline"
 #define	BL_YESNO	"yesno"
+#define	BL_SAVETTY	"savetty"
 #define	BL_ALIAS	"alias"
 #define	BL_UALIAS	"unalias"
 #define	BL_FUNCTION	"function"
@@ -458,10 +578,10 @@ extern int execusercomm __P_((char *, char *, int));
 #endif
 extern int entryhist __P_((int, char *, int));
 extern char *removehist __P_((int));
-extern int loadhistory __P_((int, char *));
+extern int loadhistory __P_((int));
 extern VOID convhistory __P_((char *, FILE *));
-extern int savehistory __P_((int, char *));
-extern int parsehist __P_((char *, int *));
+extern int savehistory __P_((int));
+extern int parsehist __P_((char *, int *, int));
 extern char *evalhistory __P_((char *));
 #ifdef	DEBUG
 extern VOID freehistory __P_((int));
@@ -496,6 +616,9 @@ extern char *newkanjiconv __P_((char *, int, int, int));
 #ifndef	_NOKANJIFCONV
 extern int getkcode __P_((char *));
 #endif
+#ifndef	_NOKANJICONV
+extern int getoutputkcode __P_((VOID_A));
+#endif
 extern char *convget __P_((char *, char *, int));
 extern char *convput __P_((char *, char *, int, int, char *, int *));
 
@@ -503,6 +626,8 @@ extern char *convput __P_((char *, char *, int, int, char *, int *));
 extern int intrkey __P_((VOID_A));
 extern int Xgetkey __P_((int, int));
 extern int kanjiputs2 __P_((char *, int, int));
+extern VOID cputspace __P_((int));
+extern VOID cputstr __P_((int, char *));
 extern char *inputstr __P_((char *, int, int, char *, int));
 extern int yesno __P_((CONST char *, ...));
 extern VOID warning __P_((int, char *));
@@ -510,6 +635,9 @@ extern int selectstr __P_((int *, int, int, char *[], int []));
 
 /* info.c */
 extern VOID help __P_((int));
+#ifndef	NOFLOCK
+extern int isnfs __P_((char *));
+#endif
 extern int writablefs __P_((char *));
 extern off_t getblocksize __P_((char *));
 extern off_t calcKB __P_((off_t, off_t));
@@ -574,6 +702,7 @@ extern int customize __P_((VOID_A));
 #endif	/* !_NOCUSTOMIZE */
 
 /* command.c */
+extern int evalstatus __P_((int));
 #ifndef	_NOSPLITWIN
 extern int nextwin __P_((VOID_A));
 #endif
@@ -584,6 +713,10 @@ extern int putmode __P_((char *, u_int, int));
 #ifdef	HAVEFLAGS
 extern int putflags __P_((char *, u_long));
 #endif
+#ifndef	NOUID
+extern int putowner __P_((char *, uid_t));
+extern int putgroup __P_((char *, gid_t));
+#endif
 extern VOID waitmes __P_((VOID_A));
 extern int filetop __P_((int));
 extern int calcwidth __P_((VOID_A));
@@ -596,4 +729,4 @@ extern VOID calcwin __P_((VOID_A));
 extern VOID movepos __P_((int, int));
 extern VOID rewritefile __P_((int));
 extern VOID addlist __P_((VOID_A));
-extern VOID main_fd __P_((char **));
+extern VOID main_fd __P_((char **, int));
