@@ -1,7 +1,7 @@
 /*
  *	kanji.c
  *
- *	Kanji Convert Function
+ *	Kanji convert functions
  */
 
 #define	K_EXTERN
@@ -47,6 +47,9 @@ extern int norealpath;
 # ifndef	_NOROCKRIDGE
 extern int norockridge;
 # endif
+# if	!defined (_NOKANJICONV) && !defined (_NOPTY)
+extern int parentfd;
+# endif
 #else	/* !FD */
 #define	realpath2(p, r, l)	realpath(p, r)
 extern char *malloc2 __P_((ALLOC_T));
@@ -63,20 +66,22 @@ extern char *includepath __P_((char *, char *));
 # else
 # define	L_SET	0
 # endif
-#endif
+#endif	/* !L_SET */
 #ifndef	L_INCR
 # ifdef	SEEK_CUR
 # define	L_INCR	SEEK_CUR
 # else
 # define	L_INCR	1
 # endif
-#endif
+#endif	/* !L_INCR */
 
 #define	ASCII		000
 #define	KANA		001
 #define	KANJI		002
 #define	JKANA		004
+#define	J_UDEF		0x222e	/* GETA */
 #define	SJ_UDEF		0x81ac	/* GETA */
+#define	U2_UDEF		0x3013	/* GETA */
 #define	UNICODETBL	"fd-unicd.tbl"
 #define	MINUNICODE	0x00a7
 #define	MAXUNICODE	0xffe5
@@ -87,6 +92,13 @@ extern char *includepath __P_((char *, char *));
 #ifndef	O_BINARY
 #define	O_BINARY	0
 #endif
+#ifdef	CODEEUC
+#define	kencode		ujis2sjis
+#define	kdecode		sjis2ujis
+#else
+#define	kencode		sjis2ujis
+#define	kdecode		ujis2sjis
+#endif
 
 typedef struct _kconv_t {
 	u_short start;
@@ -94,6 +106,14 @@ typedef struct _kconv_t {
 	u_short range;
 } kconv_t;
 
+#if	!defined (_NOKANJICONV) \
+|| (!defined (_NOENGMES) && !defined (_NOJPNMES))
+typedef struct _langtable {
+	char *ident;
+	int lang;
+} langtable;
+#endif
+
 #ifndef	_NOKANJIFCONV
 typedef struct _kpathtable {
 	char **path;
@@ -163,6 +183,34 @@ char *unitblpath = NULL;
 int unicodebuffer = 0;
 #endif
 
+#if	!defined (_NOKANJICONV) \
+|| (!defined (_NOENGMES) && !defined (_NOJPNMES))
+static CONST langtable langlist[] = {
+# ifndef	_NOKANJICONV
+	{"sjis", SJIS},
+	{"euc", EUC},
+	{"ujis", EUC},
+	{"utf8-mac", M_UTF8},
+	{"mac", M_UTF8},
+	{"utf8", UTF8},
+	{"utf-8", UTF8},
+	{"ojunet", O_JUNET},
+	{"ojis8", O_JIS8},
+	{"ojis", O_JIS7},
+	{"junet", JUNET},
+	{"jis8", JIS8},
+	{"jis", JIS7},
+	{"hex", HEX},
+	{"cap", CAP},
+# endif	/* _NOKANJICONV */
+# ifndef	_NOENGMES
+	{"eng", ENG},
+	{"C", ENG},
+# endif	/* _NOENGMES */
+};
+#define	MAXLANGLIST	arraysize(langlist)
+#endif	/* !_NOKANJICONV || (!_NOENGMES && !_NOJPNMES) */
+
 #if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
 static u_char *unitblbuf = NULL;
 static u_int unitblent = 0;
@@ -199,7 +247,7 @@ static CONST kconv_t rsjistable[] = {
 	{0xfa54, 0x81ca, 0x01},		/* full width not sign */
 	{0xfa5b, 0x81e6, 0x01},		/* because */
 };
-#define	RSJISTBLSIZ	((int)(sizeof(rsjistable) / sizeof(kconv_t)))
+#define	RSJISTBLSIZ	arraysize(rsjistable)
 #endif	/* !_NOKANJICONV || (FD && !_NODOSDRIVE) */
 #if	!defined (_NOKANJICONV) \
 || (defined (FD) && defined (_USEDOSEMU) && defined (CODEEUC))
@@ -223,7 +271,7 @@ static CONST kconv_t convtable[] = {
 	{0xfb9c, 0xee80, 0x61},		/* IBM extensions */
 	{0xfc40, 0xeee1, 0x0c}		/* IBM extensions */
 };
-#define	CNVTBLSIZ	((int)(sizeof(convtable) / sizeof(kconv_t)))
+#define	CNVTBLSIZ	arraysize(convtable)
 static CONST u_char sj2jtable1[256] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
@@ -308,6 +356,22 @@ static u_char j2sjtable2[128] = {
 	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x70 */
 	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,    0
 };
+static u_char j2sjtable3[128] = {
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
+	   0, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5,	/* 0x20 */
+	0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad,
+	0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5,	/* 0x30 */
+	0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd,
+	0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5,	/* 0x40 */
+	0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd,
+	0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5,	/* 0x50 */
+	0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd,
+	0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5,	/* 0x60 */
+	0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed,
+	0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5,	/* 0x70 */
+	0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc,    0
+};
 #endif	/* !_NOKANJICONV || (FD && _USEDOSEMU && CODEEUC) */
 #ifndef	_NOKANJICONV
 static u_int nftblnum = 0;
@@ -378,21 +442,21 @@ static CONST u_char b2htable[] = {
 #endif	/* !_NOKANJICONV */
 #ifndef	_NOKANJIFCONV
 static CONST kpathtable kpathlist[] = {
-	{ &sjispath, SJIS },
-	{ &eucpath, EUC },
-	{ &jis7path, JIS7 },
-	{ &jis8path, JIS8 },
-	{ &junetpath, JUNET },
-	{ &ojis7path, O_JIS7 },
-	{ &ojis8path, O_JIS8 },
-	{ &ojunetpath, O_JUNET },
-	{ &hexpath, HEX },
-	{ &cappath, CAP },
-	{ &utf8path, UTF8 },
-	{ &utf8macpath, M_UTF8 },
-	{ &noconvpath, NOCNV },
+	{&sjispath, SJIS},
+	{&eucpath, EUC},
+	{&jis7path, JIS7},
+	{&jis8path, JIS8},
+	{&junetpath, JUNET},
+	{&ojis7path, O_JIS7},
+	{&ojis8path, O_JIS8},
+	{&ojunetpath, O_JUNET},
+	{&hexpath, HEX},
+	{&cappath, CAP},
+	{&utf8path, UTF8},
+	{&utf8macpath, M_UTF8},
+	{&noconvpath, NOCNV},
 };
-#define	MAXKPATHLIST	(sizeof(kpathlist) / sizeof(kpathtable))
+#define	MAXKPATHLIST	arraysize(kpathlist)
 #endif	/* !_NOKANJIFCONV */
 
 #ifndef	FD
@@ -423,6 +487,9 @@ char *newkanjiconv __P_((char *, int, int, int));
 #ifndef	_NOKANJIFCONV
 int getkcode __P_((char *));
 #endif
+#ifndef	_NOKANJICONV
+int getoutputkcode __P_((VOID_A));
+#endif
 char *convget __P_((char *, char *, int));
 char *convput __P_((char *, char *, int, int, char *, int *));
 #endif	/* !FD */
@@ -445,6 +512,7 @@ int ptr;
 #endif
 	}
 	if (i > ptr) return(0);
+
 	return(iskanji1(s, i));
 }
 
@@ -458,8 +526,12 @@ char *s1, *s2;
 	while (*s1) {
 		for (i = 0;; i++) {
 			if (!s2[i]) return(s1);
-			c1 = toupper2(s1[i]);
-			c2 = toupper2(s2[i]);
+			c1 = s1[i];
+			c2 = s2[i];
+			if (islower2(c2)) {
+				c1 = toupper2(c1);
+				c2 = toupper2(c2);
+			}
 			if (c1 != c2) break;
 #ifndef	CODEEUC
 			if (issjis1(c1)) {
@@ -474,6 +546,7 @@ char *s1, *s2;
 #endif
 		s1++;
 	}
+
 	return(NULL);
 }
 
@@ -482,38 +555,21 @@ int getlang(s, io)
 char *s;
 int io;
 {
-	int ret;
+	int i, ret;
 
-	if (!s) ret = NOCNV;
-# ifndef	_NOKANJICONV
-	else if (io == L_FNAME && strstr2(s, "hex")) ret = HEX;
-	else if (io == L_FNAME && strstr2(s, "cap")) ret = CAP;
-	else if (strstr2(s, "sjis")) ret = SJIS;
-	else if (strstr2(s, "euc") || strstr2(s, "ujis")) ret = EUC;
-	else if (strstr2(s, "utf8-mac") || strstr2(s, "mac")) ret = M_UTF8;
-	else if (strstr2(s, "utf8") || strstr2(s, "utf-8")) ret = UTF8;
-	else if (strstr2(s, "ojunet")) ret = O_JUNET;
-	else if (strstr2(s, "ojis8")) ret = O_JIS8;
-	else if (strstr2(s, "ojis")) ret = O_JIS7;
-	else if (strstr2(s, "junet")) ret = JUNET;
-	else if (strstr2(s, "jis8")) ret = JIS8;
-	else if (strstr2(s, "jis")) ret = JIS7;
-# endif	/* !_NOKANJICONV */
-# ifndef	_NOENGMES
-	else if (io == L_OUTPUT && (strstr2(s, "eng") || !strcmp(s, "C")))
-		ret = ENG;
-# endif
-	else ret = NOCNV;
-
-# ifndef	_NOKANJICONV
-	if (io == L_INPUT) {
-#  ifdef	CODEEUC
-		if (ret != SJIS) ret = EUC;
-#  else
-		if (ret != EUC) ret = SJIS;
-#  endif
+	ret = NOCNV;
+	if (s) for (i = 0; i < MAXLANGLIST; i++) {
+		if (!(kanjiiomode[langlist[i].lang] & io)) continue;
+		if (strstr2(s, langlist[i].ident)) {
+			ret = langlist[i].lang;
+			break;
+		}
 	}
-# endif	/* !_NOKANJICONV */
+
+# ifndef	_NOKANJICONV
+	if (io == L_INPUT && ret == NOCNV) ret = DEFCODE;
+# endif
+
 	return(ret);
 }
 #endif	/* !_NOKANJICONV || (!_NOENGMES && !_NOJPNMES) */
@@ -532,20 +588,21 @@ static VOID NEAR sj2j(buf, s)
 char *buf;
 u_char *s;
 {
-	int s1, s2, j1, j2;
+	u_int w;
+	int i, s1, s2, j1, j2;
 
 	s1 = s[0] & 0xff;
 	s2 = s[1] & 0xff;
 	if (s1 >= 0xf0) {
-		u_int w;
-		int i;
-
 		w = (((u_int)s1 << 8) | s2);
 		for (i = 0; i < CNVTBLSIZ; i++)
 			if (w >= convtable[i].start
 			&& w < convtable[i].start + convtable[i].range)
 				break;
-		if (i >= CNVTBLSIZ) w = SJ_UDEF;
+		if (i >= CNVTBLSIZ) {
+			w = SJ_UDEF;
+			kanjierrno = SJIS;
+		}
 		else {
 			w -= convtable[i].start;
 			w += convtable[i].cnv;
@@ -555,7 +612,13 @@ u_char *s;
 	}
 	j1 = sj2jtable1[s1];
 	j2 = sj2jtable2[s2];
-	if (s2 >= 0x9f) j1++;
+	if (!j1 || !j2) {
+		j1 = ((J_UDEF >> 8) & 0xff);
+		j2 = (J_UDEF & 0xff);
+		kanjierrno = SJIS;
+	}
+	else if (s2 >= 0x9f) j1++;
+
 	buf[0] = j1;
 	buf[1] = j2;
 }
@@ -571,8 +634,12 @@ u_char *s;
 	j2 = s[1] & 0x7f;
 
 	s1 = j2sjtable1[j1];
-	s2 = (j1 & 1) ? j2sjtable2[j2] : (j2 + 0x7e);
-	w = (((u_int)s1 << 8) | s2);
+	s2 = (j1 & 1) ? j2sjtable2[j2] : j2sjtable3[j2];
+	if (!s1 || !s2) {
+		w = SJ_UDEF;
+		kanjierrno = JIS7;
+	}
+	else w = (((u_int)s1 << 8) | s2);
 
 	for (i = 0; i < CNVTBLSIZ; i++)
 	if (w >= convtable[i].cnv
@@ -607,6 +674,7 @@ int max;
 		}
 		else buf[j] = s[i];
 	}
+
 	return(j);
 }
 
@@ -623,6 +691,8 @@ int max;
 		else if (isekana(s, i)) buf[j] = s[++i];
 		else buf[j] = s[i];
 	}
+	if (kanjierrno) kanjierrno = EUC;
+
 	return(j);
 }
 #endif	/* !_NOKANJICONV || (FD && _USEDOSEMU && CODEEUC) */
@@ -655,6 +725,7 @@ char *file;
 		}
 		unitblent = getword(buf, 0);
 	}
+
 	return(fd);
 }
 
@@ -666,6 +737,7 @@ ALLOC_T size;
 	if ((tbl = (u_char *)malloc(size))) return(tbl);
 	openunitbl(NULL);
 	if (unicodebuffer) unicodebuffer = 0;
+
 	return(NULL);
 }
 
@@ -743,6 +815,7 @@ int nf;
 		nftblent = tblent;
 	}
 # endif	/* !_NOKANJICONV */
+
 	openunitbl(NULL);
 }
 
@@ -783,6 +856,7 @@ int russ;
 		wc -= rsjistable[i].start;
 		wc += rsjistable[i].cnv;
 	}
+
 	return(wc);
 }
 
@@ -802,7 +876,7 @@ int encode;
 	}
 
 	if (encode) {
-		r = 0xff00;
+		r = U2_UDEF;
 		if (wc < 0x0080) return(wc);
 		if (wc >= 0x00a1 && wc <= 0x00df)
 			return(0xff00 | (wc - 0x00a1 + 0x61));
@@ -810,11 +884,14 @@ int encode;
 			return(0xff00 | (wc - 0x8260 + 0x21));
 		if (wc >= 0x8281 && wc <= 0x829a)
 			return(0xff00 | (wc - 0x8281 + 0x41));
-		if (wc < MINKANJI || wc > MAXKANJI) return(r);
+		if (wc < MINKANJI || wc > MAXKANJI) {
+			kanjierrno = SJIS;
+			return(r);
+		}
 		wc = unifysjis(wc, 0);
 	}
 	else {
-		r = '?';
+		r = SJ_UDEF;
 		switch (wc & 0xff00) {
 			case 0:
 				if ((wc & 0xff) < 0x80) return(wc);
@@ -831,7 +908,10 @@ int encode;
 			default:
 				break;
 		}
-		if (wc < MINUNICODE || wc > MAXUNICODE) return(r);
+		if (wc < MINUNICODE || wc > MAXUNICODE) {
+			kanjierrno = UTF8;
+			return(r);
+		}
 	}
 
 	if (unicodebuffer && !unitblbuf) readunitable(0);
@@ -905,9 +985,11 @@ int encode;
 
 	if (encode) {
 		if (ofs < unitblent) r = getword(cp, 0);
+		else kanjierrno = SJIS;
 	}
 	else {
 		if (ofs > min && ofs < max) r = getword(cp, 2);
+		else kanjierrno = UTF8;
 	}
 
 	return(r);
@@ -970,6 +1052,7 @@ int max;
 # endif
 		else buf[j] = s[i];
 	}
+
 	return(j);
 }
 
@@ -1028,6 +1111,7 @@ int max, knj, asc, io;
 		buf[j++] = '(';
 		buf[j++] = asc;
 	}
+
 	return(j);
 }
 
@@ -1126,6 +1210,7 @@ int max, io;
 			else buf[j++] = s[i];
 			break;
 	}
+
 	return(j);
 }
 
@@ -1171,6 +1256,7 @@ int max, knj, asc, io;
 		buf[j++] = '(';
 		buf[j++] = asc;
 	}
+
 	return(j);
 }
 
@@ -1228,6 +1314,7 @@ int max, knj, asc, io;
 		buf[j++] = '(';
 		buf[j++] = asc;
 	}
+
 	return(j);
 }
 
@@ -1320,7 +1407,8 @@ int *ptrp, nf;
 		}
 	}
 	else if ((fd = opennftbl(UNICODETBL, nf, &ent)) < 0
-	|| (top = Xlseek(fd, (off_t)0, L_INCR)) < (off_t)0) /*EMPTY*/;
+	|| (top = Xlseek(fd, (off_t)0, L_INCR)) < (off_t)0)
+		/*EMPTY*/;
 	else {
 		n = 2 + nflen * 2;
 		cp = new = (u_char *)malloc2(n);
@@ -1358,6 +1446,7 @@ int *ptrp, nf;
 		*ptrp += i;
 	}
 	if (new) free(new);
+
 	return(w);
 }
 
@@ -1372,6 +1461,7 @@ int *ptrp;
 		u_char tmp[2];
 
 		j2sj((char *)tmp, &(s[*ptrp]));
+		if (kanjierrno) kanjierrno = EUC;
 		*ptrp += 2;
 		w = (((u_int)(tmp[0]) << 8) | tmp[1]);
 	}
@@ -1387,6 +1477,7 @@ int *ptrp;
 	else if (isskana(s, *ptrp)) w = s[(*ptrp)++];
 # endif
 	else w = s[(*ptrp)++];
+
 	return(cnvunicode(w, 1));
 }
 
@@ -1448,12 +1539,17 @@ int *ptrp;
 
 	w = s[(*ptrp)++];
 	if (w < 0x80);
-	else if ((w & 0xe0) == 0xc0 && (s[(*ptrp)++] & 0xc0) == 0x80)
+	else if (isutf2(w, s[*ptrp]))
 		w = ((w & 0x1f) << 6) | (s[(*ptrp)++] & 0x3f);
-	else {
+	else if (isutf3(w, s[*ptrp], s[*ptrp + 1])) {
 		w = ((w & 0x0f) << 6) | (s[(*ptrp)++] & 0x3f);
 		w = (w << 6) | (s[(*ptrp)++] & 0x3f);
 	}
+	else {
+		w = U2_UDEF;
+		kanjierrno = UTF8;
+	}
+
 	return(w);
 }
 
@@ -1467,6 +1563,8 @@ int max;
 	i = j = 0;
 	while (s[i] && j < max - 3) ucs2toutf8(buf, &j, toucs2(s, &i));
 	cnvunicode(0, -1);
+	if (kanjierrno) kanjierrno = DEFCODE;
+
 	return(j);
 }
 
@@ -1480,6 +1578,7 @@ int max;
 	i = j = 0;
 	while (s[i] && j < max - 2) fromucs2(buf, &j, ucs2fromutf8(s, &i));
 	cnvunicode(0, -1);
+
 	return(j);
 }
 
@@ -1504,6 +1603,8 @@ int max, nf;
 	free(u1);
 	free(u2);
 	cnvunicode(0, -1);
+	if (kanjierrno) kanjierrno = DEFCODE;
+
 	return(j);
 }
 
@@ -1528,6 +1629,7 @@ int max, nf;
 	free(u1);
 	free(u2);
 	cnvunicode(0, -1);
+
 	return(j);
 }
 
@@ -1541,6 +1643,7 @@ int c;
 	buf[i++] = ':';
 	buf[i++] = tohexa((c >> 4) & 0xf);
 	buf[i++] = tohexa(c & 0xf);
+
 	return(i);
 }
 
@@ -1570,6 +1673,8 @@ int max;
 # endif
 		else buf[j] = s[i];
 	}
+	if (kanjierrno) kanjierrno = DEFCODE;
+
 	return(j);
 }
 
@@ -1620,6 +1725,7 @@ int max;
 			else buf[j] = c1;
 		}
 	}
+
 	return(j);
 }
 
@@ -1636,6 +1742,7 @@ int c;
 		buf[i++] = tohexa((c >> 4) & 0xf);
 		buf[i++] = tohexa(c & 0xf);
 	}
+
 	return(i);
 }
 
@@ -1665,6 +1772,8 @@ int max;
 # endif
 		else buf[j] = s[i];
 	}
+	if (kanjierrno) kanjierrno = DEFCODE;
+
 	return(j);
 }
 
@@ -1715,6 +1824,7 @@ int max;
 			else buf[j] = c1;
 		}
 	}
+
 	return(j);
 }
 
@@ -1722,57 +1832,10 @@ static char *NEAR _kanjiconv(buf, s, max, in, out, lenp, io)
 char *buf, *s;
 int max, in, out, *lenp, io;
 {
+	kanjierrno = 0;
+
 	if (in == out || in == NOCNV || out == NOCNV) return(s);
 	switch (out) {
-# ifdef	CODEEUC
-		case SJIS:
-			*lenp = ujis2sjis(buf, (u_char *)s, max);
-			break;
-		case EUC:
-			switch (in) {
-				case SJIS:
-					*lenp = sjis2ujis(buf, (u_char *)s,
-						max);
-# else
-		case EUC:
-			*lenp = sjis2ujis(buf, (u_char *)s, max);
-			break;
-		case SJIS:
-			switch (in) {
-				case EUC:
-					*lenp = ujis2sjis(buf, (u_char *)s,
-						max);
-# endif
-					break;
-				case JIS7:
-				case O_JIS7:
-				case JIS8:
-				case O_JIS8:
-				case JUNET:
-				case O_JUNET:
-					*lenp = fromjis(buf, (u_char *)s,
-						max, io);
-					break;
-				case UTF8:
-					*lenp = fromutf8(buf, (u_char *)s,
-						max);
-					break;
-				case M_UTF8:
-					*lenp = fromutf8nf(buf, (u_char *)s,
-						max, NF_MAC);
-					break;
-				case HEX:
-					*lenp = fromhex(buf, (u_char *)s, max);
-					break;
-				case CAP:
-					*lenp = fromcap(buf, (u_char *)s, max);
-					break;
-				default:
-					return(s);
-/*NOTREACHED*/
-					break;
-			}
-			break;
 		case ENG:
 			*lenp = toenglish(buf, (u_char *)s, max);
 			break;
@@ -1806,11 +1869,50 @@ int max, in, out, *lenp, io;
 		case CAP:
 			*lenp = tocap(buf, (u_char *)s, max);
 			break;
+		case SECCODE:
+			*lenp = kencode(buf, (u_char *)s, max);
+			break;
+		case DEFCODE:
+			switch (in) {
+				case JIS7:
+				case O_JIS7:
+				case JIS8:
+				case O_JIS8:
+				case JUNET:
+				case O_JUNET:
+					*lenp = fromjis(buf, (u_char *)s,
+						max, io);
+					break;
+				case UTF8:
+					*lenp = fromutf8(buf, (u_char *)s,
+						max);
+					break;
+				case M_UTF8:
+					*lenp = fromutf8nf(buf, (u_char *)s,
+						max, NF_MAC);
+					break;
+				case HEX:
+					*lenp = fromhex(buf, (u_char *)s, max);
+					break;
+				case CAP:
+					*lenp = fromcap(buf, (u_char *)s, max);
+					break;
+				case SECCODE:
+					*lenp = kdecode(buf, (u_char *)s, max);
+					break;
+				default:
+					return(s);
+/*NOTREACHED*/
+					break;
+			}
+			break;
 		default:
 			return(s);
 /*NOTREACHED*/
 			break;
 	}
+	if (io == L_FNAME && kanjierrno) return(s);
+
 	return(buf);
 }
 
@@ -1822,6 +1924,8 @@ int max, in, out, io;
 
 	if (_kanjiconv(buf, s, max, in, out, &len, io) != buf)
 		for (len = 0; s[len]; len++) buf[len] = s[len];
+	buf[len] = '\0';
+
 	return(len);
 }
 
@@ -1833,6 +1937,7 @@ int max, in, out, io;
 
 	if (_kanjiconv(buf, s, max, in, out, &len, io) != buf) return(s);
 	buf[len] = '\0';
+
 	return(buf);
 }
 
@@ -1850,6 +1955,7 @@ int in, out, io;
 		free(buf);
 		return(s);
 	}
+
 	return(buf);
 }
 #endif	/* !_NOKANJICONV */
@@ -1867,10 +1973,21 @@ char *path;
 		if (includepath(path, *(kpathlist[i].path)))
 			return(kpathlist[i].code);
 	}
+
 	return(fnamekcode);
 }
 #endif	/* !_NOKANJIFCONV */
 
+#ifndef	_NOKANJICONV
+int getoutputkcode(VOID_A)
+{
+# if	defined (FD) && !defined (_NOPTY)
+	if (parentfd >= 0 && ptyoutkcode != NOCNV) return(ptyoutkcode);
+# endif
+	return(outputkcode);
+}
+#endif	/* !_NOKANJICONV */
+
 /*ARGSUSED*/
 char *convget(buf, path, dos)
 char *buf, *path;
@@ -1905,8 +2022,9 @@ int dos;
 	}
 #endif	/* FD && _USEDOSEMU */
 #ifndef	_NOKANJIFCONV
-	if (fgetok) cp = kanjiconv2(buf, cp,
-		MAXPATHLEN - 1, getkcode(cp), DEFCODE, L_FNAME);
+	if (fgetok)
+		cp = kanjiconv2(buf, cp,
+			MAXPATHLEN - 1, getkcode(cp), DEFCODE, L_FNAME);
 #endif
 #if	defined (FD) && !defined (_NOROCKRIDGE)
 	if (cp == rbuf) {
@@ -1914,6 +2032,7 @@ int dos;
 		return(buf);
 	}
 #endif
+
 	return(cp);
 }
 
@@ -1931,7 +2050,7 @@ int *codep;
 	char rbuf[MAXPATHLEN];
 #endif
 #ifndef	_NOKANJIFCONV
-	int fputok;
+	int c, fputok;
 #endif
 	char *cp, *file, rpath[MAXPATHLEN];
 	int n;
@@ -1971,7 +2090,7 @@ int *codep;
 #endif
 		else {
 			file = path;
-			strcpy(rpath, ".");
+			copycurpath(rpath);
 		}
 		realpath2(rpath, rpath, 1);
 		cp = strcatdelim(rpath);
@@ -1989,8 +2108,6 @@ int *codep;
 #endif
 #ifndef	_NOKANJIFCONV
 	if (fputok) {
-		int c;
-
 		c = getkcode(cp);
 		if (codep) *codep = c;
 		cp = kanjiconv2(kbuf, cp, MAXPATHLEN - 1, DEFCODE, c, L_FNAME);
@@ -2025,5 +2142,6 @@ int *codep;
 	else
 #endif
 	strcpy(buf, file);
+
 	return(buf);
 }
