@@ -25,6 +25,10 @@ extern char *unixrealpath __P_((char *, char *));
 #include "system.h"
 #endif
 
+#ifndef	_NOPTY
+#include "termemu.h"
+#endif
+
 #ifdef	CYGWIN
 #include <sys/cygwin.h>
 #endif
@@ -142,8 +146,9 @@ char *path, *resolved;
 int rdlink;
 {
 	char *cp, *top;
+	int n;
 
-	if (!*path || !strcmp(path, ".")) return(resolved);
+	if (!*path || (n = isdotdir(path)) == 2) return(resolved);
 	else if ((cp = strdelim(path, 0))) {
 		*cp = '\0';
 		_realpath2(path, resolved, rdlink);
@@ -152,7 +157,7 @@ int rdlink;
 		return(resolved);
 	}
 
-	if (!strcmp(path, "..")) {
+	if (n == 1) {
 		cp = strrdelim(resolved, 0);
 		top = resolved;
 #if	MSDOS
@@ -203,7 +208,7 @@ int rdlink;
 	path = tmp;
 
 #if	MSDOS
-	drv = dospath("", NULL);
+	drv = dospath(nullstr, NULL);
 	if ((resolved[0] = _dospath(path))) path += 2;
 	if (*path == _SC_) {
 		if (!resolved[0]) resolved[0] = drv;
@@ -223,7 +228,7 @@ int rdlink;
 		}
 	}
 #else	/* !MSDOS */
-	if (*path == _SC_) strcpy(resolved, _SS_);
+	if (*path == _SC_) copyrootpath(resolved);
 # ifndef	_NODOSDRIVE
 	else if ((drv = _dospath(path))) {
 		path += 2;
@@ -245,7 +250,7 @@ int rdlink;
 #endif	/* !MSDOS */
 	else if (!rdlink && resolved != fullpath && *fullpath)
 		strcpy(resolved, fullpath);
-	else if (!Xgetwd(resolved)) strcpy(resolved, _SS_);
+	else if (!Xgetwd(resolved)) copyrootpath(resolved);
 	norealpath++;
 	_realpath2(path, resolved, rdlink);
 	norealpath--;
@@ -258,17 +263,17 @@ char *path;
 {
 	char cwd[MAXPATHLEN];
 
-	if (!Xgetwd(cwd)) strcpy(cwd, _SS_);
+	if (!Xgetwd(cwd)) copyrootpath(cwd);
 	if (Xchdir(path) < 0) return(-1);
 #if	!MSDOS
 # ifndef	_NODOSDRIVE
-	if (dospath2("")) /*EMPTY*/;
+	if (dospath2(nullstr)) /*EMPTY*/;
 	else
 # endif
 	{
 		int fd, duperrno;
 
-		if ((fd = open(".", O_RDONLY, 0600)) < 0) {
+		if ((fd = open(curpath, O_RDONLY, 0600)) < 0) {
 # ifdef	CYGWIN
 			char upath[MAXPATHLEN], spath[MAXPATHLEN];
 			char tmp[MAXPATHLEN];
@@ -347,6 +352,9 @@ char *path;
 	if (unixrealpath(fullpath, tmp)) strcpy(fullpath, tmp);
 #endif
 	entryhist(1, fullpath, 1);
+#ifndef	_NOPTY
+	sendparent(TE_CHDIR, fullpath);
+#endif
 
 	return(0);
 }
@@ -361,7 +369,7 @@ int raw;
 	char *cwd;
 
 	cwd = path;
-	if (!raw && path[0] && !path[1]) switch(path[0]) {
+	if (!raw && path[0] && !path[1]) switch (path[0]) {
 		case '.':
 			cwd = NULL;
 			break;
@@ -398,7 +406,7 @@ int raw;
 		if (findpattern) free(findpattern);
 		findpattern = NULL;
 #ifndef	_NOUSEHASH
-		searchhash(NULL, "", "");
+		searchhash(NULL, nullstr, nullstr);
 #endif
 	}
 
@@ -668,7 +676,7 @@ char *s;
 {
 	int n;
 
-	if (!sscanf2(s, "%d%$", &n)) return(-1);
+	if (!sscanf2(s, "%<d%$", &n)) return(-1);
 
 	return(n);
 }
@@ -811,6 +819,9 @@ int export;
 #ifdef	_NOORIGSHELL
 	if (export) environ = _putenv2(cp, environ);
 	else environ2 = _putenv2(cp, environ2);
+# ifndef	_NOPTY
+	sendparent(TE_PUTSHELLVAR, name, value, export);
+# endif
 #else	/* !_NOORIGSHELL */
 	if (((export) ? putexportvar(cp, len) : putshellvar(cp, len)) < 0) {
 		free(cp);
@@ -859,6 +870,7 @@ int flags;
 	}
 
 	ret = dosystem(command);
+	LOG1(_LOG_NOTICE_, ret, "system(\"%s\");", command);
 #ifdef	_NOORIGSHELL
 	checkscreen(-1, -1);
 #endif
@@ -906,7 +918,7 @@ char *command;
 	sigvecset(n);
 	if (fp) {
 		if (wasttyflags & F_TTYIOMODE) {
-			Xputterms(T_KEYPAD);
+			Xputterm(T_KEYPAD);
 			Xtflush();
 		}
 	}
@@ -1071,8 +1083,7 @@ time_t t;
 	if (i > 0) {
 		i--;
 		i *= (int)sizeof(char);
-		i += (int)sizeof(struct tzhead)
-			+ ntime * (int)sizeof(char) * 4;
+		i += (int)sizeof(struct tzhead) + ntime * 4 * sizeof(char);
 		if (fseek(fp, i, 0) < 0
 		|| fread(&c, sizeof(char), 1, fp) != 1) {
 			fclose(fp);
@@ -1080,8 +1091,8 @@ time_t t;
 		}
 		i = c;
 	}
-	i *= (int)sizeof(char) * (4 + 1 + 1);
-	i += (int)sizeof(struct tzhead) + ntime * (int)sizeof(char) * (4 + 1);
+	i *= (4 + 1 + 1) * sizeof(char);
+	i += (int)sizeof(struct tzhead) + ntime * (4 + 1) * sizeof(char);
 	if (fseek(fp, i, 0) < 0
 	|| fread(buf, sizeof(char), 4, fp) != 4) {
 		fclose(fp);
@@ -1090,9 +1101,9 @@ time_t t;
 	tmp = char2long(buf);
 	tz = -tmp;
 
-	i = (int)sizeof(struct tzhead) + ntime * (int)sizeof(char) * (4 + 1)
-		+ ntype * (int)sizeof(char) * (4 + 1 + 1)
-		+ nchar * (int)sizeof(char);
+	i = (int)sizeof(struct tzhead) + ntime * (4 + 1) * sizeof(char)
+		+ ntype * (4 + 1 + 1) * sizeof(char)
+		+ nchar * sizeof(char);
 	if (fseek(fp, i, 0) < 0) {
 		fclose(fp);
 		return(tz);
