@@ -58,7 +58,7 @@ extern int maxarchive;
 #endif
 extern int wheader;
 extern char fullpath[];
-extern char *histfile;
+extern int savehist;
 extern char *helpindex[];
 extern int subwindow;
 extern int win_x;
@@ -70,6 +70,11 @@ extern char *deftmpdir;
 #if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
 extern char *unitblpath;
 #endif
+#ifndef	_NOPTY
+extern int ptymode;
+extern int ptyinternal;
+extern int parentfd;
+#endif
 
 #define	CLOCKUPDATE	10	/* sec */
 
@@ -132,13 +137,21 @@ static int xcpuerror __P_((VOID_A));
 static int xsizerror __P_((VOID_A));
 #endif
 static int NEAR minrow __P_((int));
+#ifndef	_NOEXTRAWIN
+static int NEAR biaswsize __P_((int, int));
+#endif
 #ifdef	SIGWINCH
 static int wintr __P_((VOID_A));
 #endif
-static VOID printtime __P_((int));
+static VOID NEAR printtime __P_((int));
 #ifdef	SIGALRM
 static int trapalrm __P_((VOID_A));
 #endif
+static char *NEAR getversion __P_((int *));
+#ifndef	_NOLOGGING
+static VOID NEAR startlog __P_((char **));
+static VOID NEAR endlog __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline __P_((char *, char *, int, char *));
 #endif
@@ -147,7 +160,6 @@ static int NEAR evaloption __P_((char *[]));
 static char *NEAR searchenv __P_((char *, char *[]));
 static VOID NEAR setexecname __P_((char *));
 static VOID NEAR setexecpath __P_((char *, char *[]));
-static VOID NEAR prepareexitfd __P_((VOID_A));
 
 char *origpath = NULL;
 char *progpath = NULL;
@@ -157,6 +169,9 @@ char *tmpfilename = NULL;
 int adjtty = 0;
 #endif
 int showsecond = 0;
+#if	FD >= 2
+int thruargs = 0;
+#endif
 int hideclock = 0;
 #ifdef	SIGALRM
 int noalrm = 1;
@@ -228,18 +243,21 @@ char *s;
 	if (isorgpid()) {
 		inittty(1);
 		keyflush();
-		prepareexitfd();
+		prepareexitfd(2);
 #ifndef	_NOORIGSHELL
 # ifndef	NOJOB
 		killjob();
 # endif
 		prepareexit(0);
 #endif	/* !_NOORIGSHELL */
+#ifndef	_NOPTY
+		killallpty();
+#endif
 #ifdef	DEBUG
 # if	!MSDOS
 		freeterment();
 # endif
-		VOID_C closetty();
+		closetty(&ttyio, &ttyout);
 		muntrace();
 #endif	/* DEBUG */
 	}
@@ -256,6 +274,9 @@ int sig;
 		forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 		dosallclose();
+#endif
+#ifndef	_NOLOGGING
+		endlog(sig + 128);
 #endif
 		endterm();
 		inittty(1);
@@ -266,11 +287,14 @@ int sig;
 # endif
 		prepareexit(0);
 #endif	/* !_NOORIGSHELL */
+#ifndef	_NOPTY
+		killallpty();
+#endif
 #ifdef	DEBUG
 # if	!MSDOS
 		freeterment();
 # endif
-		VOID_C closetty();
+		closetty(&ttyio, &ttyout);
 		muntrace();
 #endif	/* DEBUG */
 	}
@@ -467,6 +491,20 @@ static int xsizerror(VOID_A)
 }
 #endif
 
+VOID setlinecol(VOID_A)
+{
+	char buf[MAXLONGWIDTH + 1];
+
+	if (getconstvar("LINES")) {
+		snprintf2(buf, sizeof(buf), "%d", n_line);
+		setenv2("LINES", buf, 1);
+	}
+	if (getconstvar("COLUMNS")) {
+		snprintf2(buf, sizeof(buf), "%d", n_column);
+		setenv2("COLUMNS", buf, 1);
+	}
+}
+
 static int NEAR minrow(n)
 int n;
 {
@@ -482,6 +520,24 @@ int n;
 	return(WFILEMIN);
 }
 
+#ifndef	_NOEXTRAWIN
+static int NEAR biaswsize(n, max)
+int n, max;
+{
+	int i, row, min;
+
+	for (i = row = 0; i < n; i++) row += winvar[i].v_fileperrow + 1;
+	min = minrow(n);
+	if (max - row < min) {
+		winvar[n].v_fileperrow = min;
+		return(max - min);
+	}
+	winvar[n].v_fileperrow = max - row;
+
+	return(0);
+}
+#endif	/* !_NOEXTRAWIN */
+
 VOID checkscreen(xmax, ymax)
 int xmax, ymax;
 {
@@ -491,18 +547,27 @@ int xmax, ymax;
 #ifdef	SIGWINCH
 	nowinch++;
 #endif
-	dupn_line = n_line;
+	dupn_line = -1;
+#ifndef	_NOPTY
+	if (parentfd >= 0) row = WFILEMIN;
+	else
+#endif
 	{
 		row = wheader + WFOOTER;
+#ifdef	_NOEXTRAWIN
 		row += (minrow(win) + 1) * windows;
+#else
+		for (i = 0; i < windows; i++) row += minrow(i) + 1;
+#endif
 		row--;
 	}
 
-	saveterm(ttyio, &tty, NULL);
+	savetermio(ttyio, &tty, NULL);
 	wastty = isttyiomode;
 	Xttyiomode(1);
 	dupdumbterm = dumbterm;
 	if (xmax < 0 || ymax < 0) {
+		dupn_line = n_line;
 		xmax = WCOLUMNMIN;
 		ymax = 0;
 		dumbterm = 1;
@@ -513,7 +578,7 @@ int xmax, ymax;
 			cp = NOROW_K;
 		}
 		if (!i) {
-			Xputterm(T_CLEAR);
+			Xputterms(T_CLEAR);
 			Xlocate(0, 0);
 			keyflush();
 		}
@@ -533,12 +598,21 @@ int xmax, ymax;
 		}
 	}
 	dumbterm = dupdumbterm;
-	loadterm(ttyio, tty, NULL);
+	loadtermio(ttyio, tty, NULL);
 	isttyiomode = wastty;
 	if (tty) free(tty);
 
+	setlinecol();
 	if (n_line != dupn_line) {
+#ifdef	_NOEXTRAWIN
 		calcwin();
+#else
+		row = fileperrow(1);
+		for (i = windows - 1; i >= 0; i--) {
+			if (!(row = biaswsize(i, row))) break;
+			row--;
+		}
+#endif
 	}
 #ifdef	SIGWINCH
 	nowinch--;
@@ -550,6 +624,7 @@ VOID pollscreen(forced)
 int forced;
 {
 	static int winched = 0;
+	int x, y;
 
 	if (forced < 0) nowinch = winched = 0;
 	else if (nowinch) {
@@ -557,9 +632,11 @@ int forced;
 	}
 	else if (winched || forced) {
 		winched = 0;
+		x = n_column;
+		y = n_line;
 		checkscreen(-1, -1);
 		if (isorgpid()) {
-			rewritefile(1);
+			if (x != n_column || y != n_line) rewritefile(1);
 			if (subwindow) ungetch2(K_CTRL('L'));
 		}
 	}
@@ -579,7 +656,7 @@ static int wintr(VOID_A)
 }
 #endif	/* SIGWINCH */
 
-static VOID printtime(hide)
+static VOID NEAR printtime(hide)
 int hide;
 {
 	static time_t now;
@@ -592,6 +669,10 @@ int hide;
 		timersec = CLOCKUPDATE;
 	}
 	if (timersec-- < CLOCKUPDATE && !showsecond) return;
+#ifndef	_NOPTY
+	if (checkallpty() < 0) rewritefile(0);
+	if (isptymode() && !ptyinternal && hide <= 1) hide = 0;
+#endif
 	if (hide) return;
 
 #ifdef	DEBUG
@@ -626,6 +707,12 @@ static int trapalrm(VOID_A)
 
 	duperrno = errno;
 	signal2(SIGALRM, SIG_IGN);
+# if	!defined (_NOORIGSHELL) && !defined (_NOPTY)
+	if (isshptymode()) {
+		if (mypid == shellpid) printtime(hideclock);
+	}
+	else
+# endif
 	if (isorgpid()) printtime(hideclock);
 # ifdef	SIGWINCH
 	pollscreen(0);
@@ -683,9 +770,24 @@ int set;
 	return(old);
 }
 
-VOID title(VOID_A)
+static char *NEAR getversion(lenp)
+int *lenp;
 {
 	char *cp, *eol;
+
+	cp = strchr(version, ' ');
+	while (*(++cp) == ' ');
+	if (lenp) {
+		if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
+		*lenp = eol - cp;
+	}
+
+	return(cp);
+}
+
+VOID title(VOID_A)
+{
+	char *cp;
 	int i, len;
 
 	Xlocate(0, L_TITLE);
@@ -703,16 +805,13 @@ VOID title(VOID_A)
 	}
 	Xcputs2(" Ver.");
 	len += 5;
-	cp = strchr(version, ' ');
-	while (*(++cp) == ' ');
-	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
-	i = eol - cp;
-	Xcprintf2("%-*.*s", i, i, cp);
+	cp = getversion(&i);
+	cputstr(i, cp);
 	if (distributor) {
 		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? "" : " (c)1995-2005 T.Shirai  ";
+	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
 	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
 	while (i-- > 0) Xputch2(' ');
@@ -724,7 +823,7 @@ VOID title(VOID_A)
 #ifndef	_NOCUSTOMIZE
 VOID saveorigenviron(VOID_A)
 {
-	orighelpindex = copystrarray(NULL, helpindex, NULL, 10);
+	orighelpindex = copystrarray(NULL, helpindex, NULL, MAXHELPINDEX);
 	origbindlist = copybind(NULL, bindlist);
 # ifndef	_NOKEYMAP
 	origkeymaplist = copykeyseq(NULL);
@@ -741,6 +840,61 @@ VOID saveorigenviron(VOID_A)
 }
 #endif	/* !_NOCUSTOMIZE */
 
+#ifndef	_NOLOGGING
+static VOID NEAR startlog(argv)
+char **argv;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char *cp, *tmp, buf[MAXLOGLEN + 1];
+	int i, len;
+
+	cp = buf;
+	for (i = 1; argv && argv[i]; i++) {
+		if (cp > buf) *(cp++) = ' ';
+		tmp = killmeta(argv[i]);
+		len = snprintf2(cp, (int)sizeof(buf) - (cp - buf), "%s", tmp);
+		free(tmp);
+		if (len < 0) break;
+		cp += len;
+	}
+	*cp = '\0';
+
+	cp = getversion(&len);
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s (%-*.*s) starts; PWD=%k; ARGS=%k",
+		progname, len, len, cp, origpath, buf);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s (%-*.*s) starts; UID=%d; PWD=%k; ARGS=%k",
+		progname, len, len, cp, uid, origpath, buf);
+# endif
+}
+
+static VOID NEAR endlog(status)
+int status;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char cwd[MAXPATHLEN];
+
+	if (!Xgetwd(cwd)) strcpy(cwd, "?");
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s ends; PWD=%k; STATUS=%d",
+		progname, cwd, status);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s ends; UID=%d; PWD=%k; STATUS=%d",
+		progname, uid, cwd, status);
+# endif
+	logclose();
+}
+#endif	/* !_NOLOGGING */
+
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline(command, file, n, line)
 char *command, *file;
@@ -980,15 +1134,50 @@ char *argv, *envp[];
 	}
 }
 
-static VOID NEAR prepareexitfd(VOID_A)
+/*ARGSUSED*/
+VOID initfd(argv)
+char **argv;
+{
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	{
+#if	!MSDOS
+		if (adjtty) {
+			Xstdiomode();
+			inittty(0);
+			Xttyiomode(0);
+		}
+#endif	/* !MSDOS */
+		loadhistory(0);
+		entryhist(1, origpath, 1);
+	}
+#ifndef	_NOLOGGING
+	startlog(argv);
+#endif
+}
+
+/*ARGSUSED*/
+VOID prepareexitfd(status)
+int status;
 {
 	char cwd[MAXPATHLEN];
 
+#ifndef	_NOLOGGING
+	endlog(status);
+#endif
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	savehistory(0);
+
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
 		perror2(origpath);
 		if (!Xgetwd(cwd)) *cwd = '\0';
-		rawchdir(_SS_);
+		rawchdir(rootpath);
 	}
 	free(origpath);
 	free(progname);
@@ -1008,7 +1197,7 @@ static VOID NEAR prepareexitfd(VOID_A)
 # endif
 # ifndef	_NOCUSTOMIZE
 	if (orighelpindex) {
-		freestrarray(orighelpindex, 10);
+		freestrarray(orighelpindex, MAXHELPINDEX);
 		free(orighelpindex);
 	}
 	free(origbindlist);
@@ -1150,7 +1339,8 @@ char *argv[], *envp[];
 	}
 #endif	/* !_NOARCHIVE */
 
-	for (i = 0; i < 10; i++) helpindex[i] = strdup2(helpindex[i]);
+	for (i = 0; i < MAXHELPINDEX; i++)
+		helpindex[i] = strdup2(helpindex[i]);
 #ifdef	_USEDOSEMU
 	for (i = 0; fdtype[i].name; i++) {
 		fdtype[i].name = strdup2(fdtype[i].name);
@@ -1185,10 +1375,15 @@ char *argv[], *envp[];
 	cp = getshellname(progname, NULL, NULL);
 	if (!strpathcmp(cp, FDSHELL) || !strpathcmp(cp, "su")) {
 		i = main_shell(argc, argv, envp);
-		prepareexitfd();
+# ifndef	_NOPTY
+		killallpty();
+# endif
 		Xexit2(i);
 	}
-	fdmode = interactive = 1;
+#ifndef	_NOORIGSHELL
+	interactive =
+#endif
+	fdmode = 1;
 	setshellvar(envp);
 	prepareterm();
 	argc = initoption(argc, argv, envp);
@@ -1204,6 +1399,7 @@ char *argv[], *envp[];
 		errno = 0;
 		error(cp);
 	}
+	setlinecol();
 #ifndef	_NOCUSTOMIZE
 	saveorigenviron();
 #endif
@@ -1224,31 +1420,29 @@ char *argv[], *envp[];
 
 	i = evaloption(argv);
 	checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
-#if	!MSDOS
-	if (adjtty) {
-		Xstdiomode();
-		inittty(0);
-		Xttyiomode(0);
-	}
-#endif	/* !MSDOS */
-
-	loadhistory(0, histfile);
-	entryhist(1, origpath, 1);
+	initfd(argv);
 	Xputterms(T_CLEAR);
 
 #ifdef	SIGWINCH
 	nowinch = 0;
 #endif
-	main_fd(&(argv[i]));
+#if	FD >= 2
+	main_fd(&(argv[i]), (thruargs) ? 1 : 0);
+#else
+	main_fd(&(argv[i]), 0);
+#endif
 	sigvecset(0);
-	prepareexitfd();
+#ifndef	_NOLOGGING
+	endlog(0);
+#endif
 
 	Xstdiomode();
-#ifndef	_NOORIGSHELL
-# ifndef	NOJOB
+#if	!defined (_NOORIGSHELL) && !defined (NOJOB)
 	killjob();
-# endif
-#endif	/* !_NOORIGSHELL */
+#endif
+#ifndef	_NOPTY
+	killallpty();
+#endif
 	Xexit2(0);
 
 	return(0);
