@@ -55,10 +55,6 @@ typedef struct _kstree_t {
 } kstree_t;
 #endif	/* !MSDOS */
 
-#ifdef	USESELECTH
-#include <sys/select.h>
-#endif
-
 #ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
@@ -77,9 +73,7 @@ extern int errno;
 #include "termio.h"
 #include "term.h"
 
-#define	GETSIZE		"\033[6n"
-#define	SIZEFMT		"\033[%d;%dR"
-#define	WINTERMNAME	"iris"
+#define	WINTERMNAME		"iris"
 #ifdef	USESTRERROR
 #define	strerror2		strerror
 #else
@@ -90,7 +84,7 @@ extern char *sys_errlist[];
 #endif
 
 #ifdef	DEBUG
-extern VOID muntrace __P_ ((VOID));
+extern VOID muntrace __P_ ((VOID_A));
 extern char *_mtrace_file;
 #endif
 
@@ -105,8 +99,6 @@ typedef int	tputs_t;
 # include <term.h>
 # define	tgetnum2(s)		(s)
 # define	tgetflag2(s)		(s)
-# define	tgoto2(s, p1, p2)	tparm(s, p2, p1, \
-					0, 0, 0, 0, 0, 0, 0)
 # define	TERM_pc			pad_char
 # define	TERM_bc			NULL
 # define	TERM_co			columns
@@ -219,11 +211,9 @@ extern int tgetent __P_((char *, char *));
 extern int tgetnum __P_((char *));
 extern int tgetflag __P_((char *));
 extern char *tgetstr __P_((char *, char **));
-extern char *tgoto __P_((char *, int, int));
 extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
 # define	tgetnum2		tgetnum
 # define	tgetflag2		tgetflag
-# define	tgoto2			tgoto
 # define	TERM_pc			"pc"
 # define	TERM_bc			"bc"
 # define	TERM_co			"co"
@@ -344,16 +334,6 @@ extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
 #endif
 #endif	/* !MSDOS */
 
-#ifndef	FD_SET
-typedef struct fd_set {
-	u_int fds_bits[1];
-} fd_set;
-# define	FD_ZERO(p)	(((p) -> fds_bits[0]) = 0)
-# define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= ((u_int)1 << (n)))
-#endif	/* !FD_SET */
-
-#define	MAXFDSET	256
-
 static int NEAR err2 __P_((char *));
 #if	!MSDOS
 static char *NEAR tstrdup __P_((char *));
@@ -383,7 +363,7 @@ static char *NEAR tgetstr3 __P_((char **, char *, char *));
 static char *NEAR tgetkeyseq __P_((int, char *));
 static kstree_t *NEAR newkeyseqtree __P_((kstree_t *, int));
 static int NEAR freekeyseqtree __P_((kstree_t *, int));
-static int NEAR cmpkeyseq __P_((CONST VOID_P, CONST VOID_P));
+static int cmpkeyseq __P_((CONST VOID_P, CONST VOID_P));
 static int NEAR sortkeyseq __P_((VOID_A));
 static int putch3 __P_((tputs_t));
 #endif	/* !MSDOS */
@@ -417,6 +397,7 @@ int (*keywaitfunc)__P_((VOID_A)) = NULL;
 #if	!MSDOS
 int usegetcursor = 0;
 int suspended = 0;
+char *duptty[2] = {NULL, NULL};
 #endif
 int ttyio = -1;
 int isttyiomode = 0;
@@ -465,10 +446,10 @@ static int specialkeycode[] = {
 	K_F(11), K_F(12), K_F(13), K_F(14), K_F(15),
 	K_F(16), K_F(17), K_F(18), K_F(19), K_F(20), K_HELP
 };
-#define	SPECIALKEYSIZ	((int)(sizeof(specialkeycode) / sizeof(int)))
+#define	SPECIALKEYSIZ	arraysize(specialkeycode)
 #else	/* !MSDOS */
 static char *dumblist[] = {"dumb", "un", "unknown"};
-#define	DUMBLISTSIZE	((int)sizeof(dumblist) / sizeof(char *))
+#define	DUMBLISTSIZE	arraysize(dumblist)
 static keyseq_t keyseq[K_MAX - K_MIN + 1];
 static kstree_t *keyseqtree = NULL;
 static char *defkeyseq[K_MAX - K_MIN + 1] = {
@@ -477,7 +458,7 @@ static char *defkeyseq[K_MAX - K_MIN + 1] = {
 	"\033OA",		/* K_UP */
 	"\033OD",		/* K_LEFT */
 	"\033OC",		/* K_RIGHT */
-	"\033[4~",		/* K_HOME */
+	"\033[1~",		/* K_HOME */
 	"\b",			/* K_BS */
 	NULL,			/* K_F0 */
 	"\033[11~",		/* K_F(1) */
@@ -560,8 +541,8 @@ static char *defkeyseq[K_MAX - K_MIN + 1] = {
 	NULL,			/* K_EOL */
 	NULL,			/* K_ESF */
 	NULL,			/* K_ESR */
-	"\033[5~",		/* K_PPAGE */
 	"\033[6~",		/* K_NPAGE */
+	"\033[5~",		/* K_PPAGE */
 	NULL,			/* K_STAB */
 	NULL,			/* K_CTAB */
 	NULL,			/* K_CATAB */
@@ -582,7 +563,7 @@ static char *defkeyseq[K_MAX - K_MIN + 1] = {
 	NULL,			/* K_COMM */
 	NULL,			/* K_COPY */
 	NULL,			/* K_CREAT */
-	"\033[1~",		/* K_END */
+	"\033[4~",		/* K_END */
 	NULL,			/* K_EXIT */
 	NULL,			/* K_FIND */
 	NULL,			/* K_HELP */
@@ -613,8 +594,8 @@ static char *deftermstr[MAXTERMSTR] = {
 	"",			/* T_METAMODE */
 	"",			/* T_NOMETAMODE */
 #endif
-#if	MSDOS
 	"",			/* T_SCROLL */
+#if	MSDOS
 	"",			/* T_KEYPAD */
 	"",			/* T_NOKEYPAD */
 	"\033[>5l",		/* T_NORMALCURSOR */
@@ -623,11 +604,6 @@ static char *deftermstr[MAXTERMSTR] = {
 	"\033[s",		/* T_SETCURSOR */
 	"\033[u",		/* T_RESETCURSOR */
 #else	/* !MSDOS */
-# ifdef	USETERMINFO
-	"\033[%i%p1%d;%p2%dr",	/* T_SCROLL */
-# else
-	"\033[%i%d;%dr",	/* T_SCROLL */
-# endif
 # ifdef	BOW
 	/* hack for bowpad */
 	"",			/* T_KEYPAD */
@@ -666,7 +642,7 @@ static char *deftermstr[MAXTERMSTR] = {
 #if	MSDOS
 	"\033[%d;%dH",		/* C_LOCATE */
 #else
-# if	defined (LINUX) || defined (USETERMINFO)
+# ifdef	USETERMINFO
 	"\033[%i%p1%d;%p2%dH",	/* C_LOCATE */
 # else
 	"\033[%i%d;%dH",	/* C_LOCATE */
@@ -681,7 +657,7 @@ static char *deftermstr[MAXTERMSTR] = {
 	"\n",			/* C_DOWN */
 	"\033[C",		/* C_RIGHT */
 	"\b",			/* C_LEFT */
-#if	!MSDOS && defined (USETERMINFO)
+#ifdef	USETERMINFO
 	"\033[%p1%dA",		/* C_NUP */
 	"\033[%p1%dB",		/* C_NDOWN */
 	"\033[%p1%dC",		/* C_NRIGHT */
@@ -977,8 +953,7 @@ int cooked2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(0, CBREAK | RAW, LPASS8, LLITOUT | LPENDIN);
 #else
-	ttymode(ISIG | ICANON | IEXTEN, PENDIN,
-		BRKINT | IXON, IGNBRK | ISTRIP,
+	ttymode(TIO_LCOOKED, PENDIN, TIO_ICOOKED, ~TIO_INOCOOKED,
 		OPOST, 0, VAL_VMIN, VAL_VTIME);
 #endif
 
@@ -990,7 +965,7 @@ int cbreak2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(CBREAK, 0, LLITOUT, 0);
 #else
-	ttymode(ISIG | IEXTEN, ICANON, BRKINT | IXON, IGNBRK, OPOST, 0, 1, 0);
+	ttymode(TIO_LCBREAK, ICANON, TIO_ICOOKED, IGNBRK, OPOST, 0, 1, 0);
 #endif
 
 	return(0);
@@ -1001,8 +976,7 @@ int raw2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(RAW, 0, LLITOUT, 0);
 #else
-	ttymode(0, ISIG | ICANON | IEXTEN,
-		IGNBRK, BRKINT | IXON, 0, OPOST, 1, 0);
+	ttymode(0, TIO_LCOOKED, IGNBRK, TIO_ICOOKED, 0, OPOST, 1, 0);
 #endif
 
 	return(0);
@@ -1013,7 +987,7 @@ int echo2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(ECHO, 0, LCRTBS | LCRTERA | LCRTKIL | LCTLECH, 0);
 #else
-	ttymode(ECHO | ECHOE | ECHOCTL | ECHOKE, ECHONL, 0, 0, 0, 0, 0, 0);
+	ttymode(TIO_LECHO, ECHONL, 0, 0, 0, 0, 0, 0);
 #endif
 
 	return(0);
@@ -1024,7 +998,7 @@ int noecho2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(0, ECHO, 0, LCRTBS | LCRTERA);
 #else
-	ttymode(0, ECHO | ECHOE | ECHOK | ECHONL, 0, 0, 0, 0, 0, 0);
+	ttymode(0, ~TIO_LNOECHO, 0, 0, 0, 0, 0, 0);
 #endif
 
 	return(0);
@@ -1035,7 +1009,7 @@ int nl2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(CRMOD, 0, 0, 0);
 #else
-	ttymode(0, 0, ICRNL, 0, ONLCR, OCRNL | ONOCR | ONLRET, 0, 0);
+	ttymode(0, 0, ICRNL, 0, ONLCR, ~TIO_ONONL, 0, 0);
 #endif
 
 	return(0);
@@ -1090,30 +1064,56 @@ int keyflush(VOID_A)
 
 	return(0);
 }
+
+int savettyio(reset)
+int reset;
+{
+	char *tty;
+	int n;
+
+	if (reset) tty = NULL;
+	else {
+		savetermio(ttyio, &tty, NULL);
+		if (!tty) return(-1);
+	}
+
+	n = (isttyiomode) ? 0 : 1;
+	if (duptty[n]) free(duptty[n]);
+	duptty[n] = tty;
+
+	return(n);
+}
 #endif	/* !MSDOS */
 
 int ttyiomode(isnl)
 int isnl;
 {
+	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[0]) loadtermio(ttyio, duptty[0], NULL);
+#endif
+	else {
 #if	MSDOS
-	raw2();
+		raw2();
 #else	/* !MSDOS */
 # ifdef	USESGTTY
-	raw2();
-	noecho2();
-	nonl2();
-	notabs();
+		raw2();
+		noecho2();
+		nonl2();
+		notabs();
+		if (dumbterm > 1) ttymode(ECHO | CRMOD, 0, 0, 0);
 # else	/* !USESGTTY */
-	if (isnl) ttymode(0, (ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOK|ECHONL),
-		IGNBRK, (BRKINT|IXON) | ICRNL,
-		OPOST | ONLCR | TAB3, 0, 1, 0);
-	else ttymode(0, (ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOK|ECHONL),
-		IGNBRK, (BRKINT|IXON) | ICRNL,
-		TAB3, OPOST | ONLCR, 1, 0);
+		if (isnl) ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
+			IGNBRK, TIO_ICOOKED | ICRNL, TIO_ONL | TAB3, 0, 1, 0);
+		else ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
+			IGNBRK, TIO_ICOOKED | ICRNL, TAB3, TIO_ONL, 1, 0);
+		if (dumbterm > 1)
+			ttymode(ECHO, 0, ICRNL, 0, TIO_ONL, 0, 1, 0);
 # endif	/* !USESGTTY */
 #endif	/* !MSDOS */
+	}
 	if (!dumbterm) {
-		putterms(T_KEYPAD);
+		putterm(T_KEYPAD);
 		tflush();
 	}
 	isttyiomode = isnl + 1;
@@ -1126,32 +1126,35 @@ int stdiomode(VOID_A)
 	int isnl;
 
 	isnl = (isttyiomode) ? isttyiomode - 1 : 0;
+	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[1]) loadtermio(ttyio, duptty[1], NULL);
+#endif
+	else {
 #if	MSDOS
-	cooked2();
+		cooked2();
 #else	/* !MSDOS */
 # ifdef	USESGTTY
-	cooked2();
-	echo2();
-	nl2();
-	tabs();
-	if (dumbterm > 2) ttymode(0, ECHO | CRMOD, 0, 0);
+		cooked2();
+		echo2();
+		nl2();
+		tabs();
+		if (dumbterm > 2) ttymode(0, ECHO | CRMOD, 0, 0);
 # else	/* !USESGTTY */
-	if (isnl) ttymode((ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOCTL|ECHOKE),
-		PENDIN | ECHONL,
-		(BRKINT|IXON) | ICRNL, (IGNBRK|ISTRIP),
-		0, (OCRNL|ONOCR|ONLRET) | TAB3,
-		VAL_VMIN, VAL_VTIME);
-	else ttymode((ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOCTL|ECHOKE),
-		PENDIN | ECHONL,
-		(BRKINT|IXON) | ICRNL, (IGNBRK|ISTRIP),
-		OPOST | ONLCR, (OCRNL|ONOCR|ONLRET) | TAB3,
-		VAL_VMIN, VAL_VTIME);
-	if (dumbterm > 2)
-		ttymode(0, ECHO, 0, ICRNL, 0, ONLCR, VAL_VMIN, VAL_VTIME);
+		if (isnl) ttymode(TIO_LCOOKED | TIO_LECHO, PENDIN | ECHONL,
+			TIO_ICOOKED | ICRNL, ~TIO_INOCOOKED,
+			0, ~TIO_ONONL | TAB3, VAL_VMIN, VAL_VTIME);
+		else ttymode(TIO_LCOOKED | TIO_LECHO, PENDIN | ECHONL,
+			TIO_ICOOKED | ICRNL, ~TIO_INOCOOKED,
+			TIO_ONL, ~TIO_ONONL | TAB3, VAL_VMIN, VAL_VTIME);
+		if (dumbterm > 2)
+			ttymode(0, ECHO,
+				0, ICRNL, 0, ONLCR, VAL_VMIN, VAL_VTIME);
 # endif	/* !USESGTTY */
 #endif	/* !MSDOS */
+	}
 	if (!dumbterm) {
-		putterms(T_NOKEYPAD);
+		putterm(T_NOKEYPAD);
 		tflush();
 	}
 	isttyiomode = 0;
@@ -1167,8 +1170,10 @@ int init;
 
 	oldmode = mode;
 	if (init >= 0 && mode != init) {
-		putterms((init) ? T_INIT : T_END);
-		tflush();
+		if (!dumbterm) {
+			putterms((init) ? T_INIT : T_END);
+			tflush();
+		}
 		mode = init;
 	}
 
@@ -1320,7 +1325,7 @@ int *yp, *xp;
 # if	!MSDOS
 	if (!usegetcursor) return(-1);
 # endif
-	if (getxy(&x, &y) < 0) x = y = -1;
+	if (getxy(&x, &y) < 0) x = y = 0;
 # if	MSDOS
 	if ((cp = tparamstr(termstr[C_LOCATE], 0, 999))) {
 		for (i = 0; cp[i]; i++) bdos(0x06, cp[i], 0);
@@ -1335,7 +1340,7 @@ int *yp, *xp;
 	tflush();
 # endif
 	i = getxy(xp, yp);
-	if (x >= 0 && y >= 0) locate(x, y);
+	if (x > 0 && y > 0) locate(--x, --y);
 
 	return(i);
 }
@@ -1343,27 +1348,24 @@ int *yp, *xp;
 int getxy(xp, yp)
 int *xp, *yp;
 {
-# if	!MSDOS
-	char *tty;
-# endif
-	char *format, buf[sizeof(SIZEFMT) + 4];
+	char *format, buf[sizeof(SIZEFMT) + (MAXLONGWIDTH - 2) * 2];
 	int i, j, tmp, count, *val[2];
 
 	format = SIZEFMT;
 	keyflush();
 # if	MSDOS
-	for (i = 0; i < (int)sizeof(GETSIZE) - 1; i++)
+	for (i = 0; i < strsize(GETSIZE); i++)
 		bdos(0x06, GETSIZE[i], 0);
 # else
 	if (!usegetcursor) return(-1);
-	saveterm(ttyio, &tty, NULL);
-	noecho2();
-	write(ttyio, GETSIZE, sizeof(GETSIZE) - 1);
+	tputs2(GETSIZE, 1);
+	tflush();
 # endif
 
 	i = 0;
+	buf[i] = '\0';
 	do {
-		if (!kbhit2(WAITKEYPAD * 1000L)) break;
+		if (!kbhit2(WAITKEYPAD * 1000L * 2)) break;
 # if	MSDOS
 		buf[i] = bdos(0x07, 0x00, 0);
 # else
@@ -1372,7 +1374,7 @@ int *xp, *yp;
 # endif
 	} while (buf[i] != format[0]);
 
-	if (buf[i] == format[0]) while (i < (int)sizeof(buf) - 2) {
+	if (buf[i] == format[0]) while (i < strsize(buf) - 1) {
 		if (!kbhit2(WAITKEYPAD * 1000L)) break;
 # if	MSDOS
 		buf[++i] = bdos(0x07, 0x00, 0);
@@ -1380,14 +1382,10 @@ int *xp, *yp;
 		if ((tmp = getch2()) == EOF) break;
 		buf[++i] = tmp;
 # endif
-		if (buf[i] == format[sizeof(SIZEFMT) - 2]) break;
+		if (buf[i] == format[strsize(SIZEFMT) - 1]) break;
 	}
 	keyflush();
-# if	!MSDOS
-	loadterm(ttyio, tty, NULL);
-	if (tty) free(tty);
-# endif
-	if (!i || buf[i] != format[sizeof(SIZEFMT) - 2]) return(-1);
+	if (!i || buf[i] != format[strsize(SIZEFMT) - 1]) return(-1);
 	buf[++i] = '\0';
 
 	count = 0;
@@ -1568,8 +1566,18 @@ char *s;
 	char *p;
 
 	p = strbuf;
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
 	s = tgetstr(s, &p);
-# endif
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+#  else
+	s = tgetstr(s, &p);
+#  endif
+# endif	/* !USETERMINFO */
 	if (s) {
 		if (*term) free(*term);
 		*term = tstrdup(s);
@@ -1584,14 +1592,35 @@ char **term, *str1, *str2;
 # ifdef	USETERMINFO
 	if (str1) str1 = tstrdup(str1);
 	else str1 = tparamstr(str2, 1, 1);
-# else
+# else	/* !USETERMINFO */
 	char *p, strbuf[TERMCAPSIZE];
 
 	p = strbuf;
 
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
+	str1 = tgetstr(str1, &p);
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+	if (str1) str1 = tstrdup(str1);
+	else {
+		_mtrace_file = "tgetstr(start)";
+		str2 = tgetstr(str2, &p);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "tgetstr(end)";
+			malloc(0);	/* dummy alloc */
+		}
+		str1 = tparamstr(str2, 1, 1);
+	}
+#  else
 	if ((str1 = tgetstr(str1, &p))) str1 = tstrdup(str1);
 	else str1 = tparamstr(tgetstr(str2, &p), 1, 1);
-# endif
+#  endif
+# endif	/* !USETERMINFO */
 	if (str1) {
 		if (*term) free(*term);
 		*term = str1;
@@ -1637,8 +1666,8 @@ int num;
 	int i, n;
 
 	if (!parent || !(parent -> next))
-		new = (kstree_t *)malloc(sizeof(kstree_t) * num);
-	else new = (kstree_t *)realloc(parent -> next, sizeof(kstree_t) * num);
+		new = (kstree_t *)malloc(num * sizeof(kstree_t));
+	else new = (kstree_t *)realloc(parent -> next, num * sizeof(kstree_t));
 	if (!new) err2("realloc()");
 
 	if (!parent) n = 0;
@@ -1670,7 +1699,7 @@ int n;
 	return(0);
 }
 
-static int NEAR cmpkeyseq(vp1, vp2)
+static int cmpkeyseq(vp1, vp2)
 CONST VOID_P vp1;
 CONST VOID_P vp2;
 {
@@ -1745,7 +1774,7 @@ char *s;
 		if (i < DUMBLISTSIZE) dumbterm = 1;
 	}
 # ifdef	IRIX
-	winterm = !strncmp(term, WINTERMNAME, sizeof(WINTERMNAME) - 1);
+	winterm = !strncmp(term, WINTERMNAME, strsize(WINTERMNAME));
 # endif
 
 	for (;;) {
@@ -1902,7 +1931,7 @@ char *s;
 	/* Hack for HP-UX 10.20 */
 	cp = NULL;
 	if (tgetstr2(&cp, TERM_AB) || tgetstr2(&cp, TERM_Sb)) {
-		if (termstr[T_FGCOLOR]) free(termstr[T_BGCOLOR]);
+		if (termstr[T_BGCOLOR]) free(termstr[T_BGCOLOR]);
 		termstr[T_BGCOLOR] = cp;
 	}
 # endif
@@ -2167,9 +2196,9 @@ keyseq_t *list;
 int initterm(VOID_A)
 {
 	if (!(termflags & F_TERMENT)) getterment(NULL);
+	termmode(1);
 	if (!dumbterm) {
-		putterms(T_KEYPAD);
-		termmode(1);
+		putterm(T_KEYPAD);
 		tflush();
 	}
 	termflags |= F_INITTERM;
@@ -2180,9 +2209,9 @@ int initterm(VOID_A)
 int endterm(VOID_A)
 {
 	if (!(termflags & F_INITTERM)) return(-1);
+	termmode(0);
 	if (!dumbterm) {
-		putterms(T_NOKEYPAD);
-		termmode(0);
+		putterm(T_NOKEYPAD);
 		tflush();
 	}
 	termflags &= ~F_INITTERM;
@@ -2190,6 +2219,15 @@ int endterm(VOID_A)
 	return(0);
 }
 
+int putterm(n)
+int n;
+{
+	if (n < 0 || n >= MAXTERMSTR) return(-1);
+	if (!termstr[n]) return(0);
+
+	return(tputs2(termstr[n], 1));
+}
+
 #if	MSDOS
 # ifdef	USEVIDEOBIOS
 static int NEAR bioslocate(x, y)
@@ -2573,37 +2611,27 @@ char *s;
 }
 # endif	/* !USEVIDEOBIOS */
 
-int putterm(n)
-int n;
-{
-	if (n < 0 || n >= MAXTERMSTR) return(-1);
-	if (!termstr[n]) return(0);
-
-	return(cputs2(termstr[n]));
-}
-
 /*ARGSUSED*/
 int kbhit2(usec)
 long usec;
 {
 #if	defined (NOTUSEBIOS) || !defined (DJGPP) || (DJGPP < 2)
 	union REGS reg;
-#else
+#else	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 # ifndef	PC98
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 # endif
-#endif
+#endif	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 
 	if (ungetnum > 0) return(1);
 #ifdef	NOTUSEBIOS
 	if (nextchar) return(1);
 	reg.x.ax = 0x4406;
 	reg.x.bx = ttyio;
-	putterms(T_METAMODE);
+	putterm(T_METAMODE);
 	int86(0x21, &reg, &reg);
-	putterms(T_NOMETAMODE);
+	putterm(T_NOMETAMODE);
 
 	return((reg.x.flags & 1) ? 0 : reg.h.al);
 #else	/* !NOTUSEBIOS */
@@ -2615,17 +2643,9 @@ long usec;
 	return(reg.h.bh != 0);
 # else	/* !PC98 */
 #  if	defined (DJGPP) && (DJGPP >= 2)
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 #  else	/* !DJGPP || DJGPP < 2 */
@@ -2690,16 +2710,16 @@ int getch2(VOID_A)
 		if (strchr(specialkey, key >> 8)) break;
 		if ((top += 2) >= KEYBUFWORKMAX) top = KEYBUFWORKMIN;
 	}
-	putterms(T_METAMODE);
+	putterm(T_METAMODE);
 	ch = (bdos(0x07, 0x00, 0) & 0xff);
-	putterms(T_NOMETAMODE);
+	putterm(T_NOMETAMODE);
 	keybuftop = getkeybuf(KEYBUFWORKTOP);
 	if (!(key & 0xff)) {
 		while (kbhit2(1000000L / SENSEPERSEC)) {
 			if (keybuftop != getkeybuf(KEYBUFWORKTOP)) break;
-			putterms(T_METAMODE);
+			putterm(T_METAMODE);
 			bdos(0x07, 0x00, 0);
-			putterms(T_NOMETAMODE);
+			putterm(T_NOMETAMODE);
 		}
 		ch = '\0';
 		nextchar = (key >> 8);
@@ -2748,7 +2768,7 @@ int sig;
 			raise(sig);
 		}
 # endif
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	} while (!i);
 
 #else	/* !DJGPP || NOTUSEBIOS || PC98 */
@@ -2761,7 +2781,7 @@ int sig;
 # endif
 			memcpy((char *)tbuf1, (char *)tbuf2, sizeof(tbuf1));
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	}
 #endif	/* !DJGPP || NOTUSEBIOS || PC98 */
 	if ((ch = getch2()) == EOF) return(K_NOKEY);
@@ -2795,9 +2815,12 @@ int sig;
 int ungetch2(c)
 int c;
 {
-	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
-		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 
 	return(c);
 }
@@ -2873,13 +2896,11 @@ tputs_t c;
 	return(fputc(c & 0x7f, ttyout));
 }
 
-int putterm(n)
+int tputs2(s, n)
+char *s;
 int n;
 {
-	if (n < 0 || n >= MAXTERMSTR) return(-1);
-	if (!termstr[n]) return(0);
-
-	return(tputs(termstr[n], 1, putch3));
+	return(tputs(s, n, putch3));
 }
 
 int putterms(n)
@@ -2888,7 +2909,7 @@ int n;
 	if (n < 0 || n >= MAXTERMSTR) return(-1);
 	if (!termstr[n]) return(0);
 
-	return(tputs(termstr[n], n_line, putch3));
+	return(tputs2(termstr[n], n_line));
 }
 
 int kbhit2(usec)
@@ -2897,21 +2918,12 @@ long usec;
 # ifdef	NOSELECT
 	return((usec) ? 1 : 0);
 # else
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 # endif
@@ -2950,7 +2962,7 @@ int sig;
 			ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
 			suspended = 0;
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 # ifndef	TIOCSTI
 		if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
 # endif
@@ -2960,8 +2972,12 @@ int sig;
 # if	!defined (_NOKANJICONV) || defined (CODEEUC)
 	else if (key != C_EKANA) /*EMPTY*/;
 #  if	!defined (_NOKANJICONV)
+#   ifdef	CODEEUC
+	else if (inputkcode != EUC && inputkcode != NOCNV) /*EMPTY*/;
+#   else
 	else if (inputkcode != EUC) /*EMPTY*/;
-#  endif
+#   endif
+#  endif	/* !_NOKANJICONV */
 	else {
 		if (!kbhit2(WAITMETA * 1000L) || (ch = getch2()) == EOF)
 			return(key);
@@ -2981,10 +2997,12 @@ int sig;
 			return(mkmetakey(ch));
 		for (j = 0; j < p -> num; j++)
 			if (key == keyseq[p -> next[j].key].str[0]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) {
+			ungetch2(ch);
+			return(key);
+		}
 		p = &(p -> next[j]);
-		if (keyseq[p -> key].len == 1)
-			return(keyseq[p -> key].code);
+		if (keyseq[p -> key].len == 1) return(keyseq[p -> key].code);
 	}
 	else {
 		for (j = 0; j < p -> num; j++)
@@ -3000,14 +3018,19 @@ int sig;
 	for (i = 1; p && p -> next; i++) {
 		for (j = 0; j < p -> num; j++)
 			if (ch == keyseq[p -> next[j].key].str[i]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) break;
 		p = &(p -> next[j]);
 		if (keyseq[p -> key].len == i + 1)
 			return(keyseq[p -> key].code);
 		if (!kbhit2(WAITKEYPAD * 1000L) || (ch = getch2()) == EOF)
-			return(key);
+			break;
 	}
 
+	for (j = 1; j < i; j++) {
+		if (j >= keyseq[p -> key].len) break;
+		ungetch2(keyseq[p -> key].str[j]);
+	}
+	ungetch2(ch);
 	return(key);
 }
 
@@ -3016,13 +3039,13 @@ int sig;
 {
 	int ch;
 
-	ch = getkey2(sig);
+	if ((ch = getkey2(sig)) < 0) return(ch);
 	if (ch >= K_F('*') && ch < K_DL) {
 		if (ch == K_F('?')) ch = K_CR;
 		else ch -= K_F0;
 	}
 
-	return(ch & 0777);
+	return(ch & 01777);
 }
 
 int ungetch2(c)
@@ -3034,9 +3057,12 @@ int c;
 	ch = c;
 	Xioctl(ttyio, TIOCSTI, &ch);
 # else
-	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
-		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 # endif
 
 	return(c);
@@ -3047,10 +3073,12 @@ int s, e;
 {
 	char *cp;
 
-	if ((cp = tparamstr(termstr[T_SCROLL], s, e))) {
-		tputs(cp, n_line, putch3);
-		free(cp);
+	if (!(cp = tparamstr(termstr[T_SCROLL], s, e)) || !*cp) {
+		if (cp) free(cp);
+		return(-1);
 	}
+	tputs2(cp, 1);
+	free(cp);
 
 	return(0);
 }
@@ -3058,20 +3086,14 @@ int s, e;
 int locate(x, y)
 int x, y;
 {
-# ifdef	DEBUG
 	char *cp;
 
-	_mtrace_file = "tgoto(start)";
-	cp = tgoto2(termstr[C_LOCATE], x, y);
-	if (_mtrace_file) _mtrace_file = NULL;
-	else {
-		_mtrace_file = "tgoto(end)";
-		malloc(0);	/* dummy malloc */
+	if (!(cp = tparamstr(termstr[C_LOCATE], y, x)) || !*cp) {
+		if (cp) free(cp);
+		return(-1);
 	}
-	tputs(cp, n_line, putch3);
-# else
-	tputs(tgoto2(termstr[C_LOCATE], x, y), n_line, putch3);
-# endif
+	tputs2(cp, 1);
+	free(cp);
 
 	return(0);
 }
@@ -3116,10 +3138,11 @@ int xmax, ymax;
 
 	if (dumbterm) /*EMPTY*/;
 	else if (usegetcursor || x < 0 || y < 0) {
-		setscroll(-1, -1);
-		if (maxlocate(&ty, &tx) >= 0) {
-			x = tx;
-			y = ty;
+		if (usegetcursor) VOID_C setscroll(-1, -1);
+		else VOID_C setscroll(0, 998);
+		if (maxlocate(&ty, &tx) >= 0 && (tx > x || ty > y)) {
+			if (tx > x) x = tx;
+			if (ty > y) y = ty;
 			VOID_C setwsize(ttyio, x, y);
 		}
 	}
@@ -3135,7 +3158,7 @@ int xmax, ymax;
 	if (n_line <= 0 || (ymax > 0 && n_line < ymax))
 		return("Line size too small");
 
-	if (xmax > 0 && ymax > 0) setscroll(-1, n_line - 1);
+	if (xmax > 0 && ymax > 0) VOID_C setscroll(0, n_line - 1);
 
 	return(NULL);
 }
@@ -3191,10 +3214,10 @@ va_dcl
 	int n;
 
 	VA_START(args, fmt);
-
 	n = vasprintf2(&buf, fmt, args);
 	va_end(args);
 	if (n < 0) err2("malloc()");
+
 	cputs2(buf);
 	free(buf);
 
@@ -3231,14 +3254,14 @@ int color, reverse;
 	}
 
 	if ((cp = tparamstr(termstr[T_FGCOLOR], fg, 0))) {
-		cputs2(cp);
+		tputs2(cp, 1);
 		free(cp);
 	}
 	else cprintf2("\033[%dm", fg + ANSI_NORMAL);
 
 	if (bg < 0) /*EMPTY*/;
 	else if ((cp = tparamstr(termstr[T_BGCOLOR], bg, 0))) {
-		cputs2(cp);
+		tputs2(cp, 1);
 		free(cp);
 	}
 	else cprintf2("\033[%dm", bg + ANSI_REVERSE);
@@ -3252,13 +3275,9 @@ int n1, n2, c;
 	char *cp;
 
 	if (n1 < 0 || !termstr[n1] || !(cp = tparamstr(termstr[n1], c, c)))
-		while (c--) putterm(n2);
+		while (c--) putterms(n2);
 	else {
-#if	MSDOS
-		cputs2(cp);
-#else
-		tputs(cp, 1, putch3);
-#endif
+		tputs2(cp, n_line);
 		free(cp);
 	}
 
