@@ -19,6 +19,9 @@ typedef struct _keyseq_t {
 	char *str;
 } keyseq_t;
 
+#define	GETSIZE	"\033[6n"
+#define	SIZEFMT	"\033[%d;%dR"
+
 #define	K_CR	'\r'
 #define	K_ESC	'\033'
 
@@ -148,6 +151,7 @@ extern int (*keywaitfunc)__P_((VOID_A));
 #if	!MSDOS
 extern int usegetcursor;
 extern int suspended;
+extern char *duptty[2];
 #endif
 extern int ttyio;
 extern int isttyiomode;
@@ -165,6 +169,9 @@ extern int nonl2 __P_((VOID_A));
 extern int tabs __P_((VOID_A));
 extern int notabs __P_((VOID_A));
 extern int keyflush __P_((VOID_A));
+#if	!MSDOS
+extern int savettyio __P_((int));
+#endif
 extern int ttyiomode __P_((int));
 extern int stdiomode __P_((VOID_A));
 extern int termmode __P_((int));
@@ -184,12 +191,14 @@ extern int freekeyseq __P_((keyseq_t *));
 #endif
 extern int initterm __P_((VOID_A));
 extern int endterm __P_((VOID_A));
+extern int putterm __P_((int));
 extern int putch2 __P_((int));
 extern int cputs2 __P_((char *));
-extern int putterm __P_((int));
 #if	MSDOS
+#define	tputs2(s,n)	cputs2(s)
 #define	putterms	putterm
 #else
+extern int tputs2 __P_((char *, int));
 extern int putterms __P_((int));
 #endif
 extern int kbhit2 __P_((long));
@@ -233,4 +242,4 @@ extern int movecursor __P_((int, int, int));
 #define	ANSI_NORMAL	30
 #define	ANSI_REVERSE	40
 
-#endif	/* __TERM_H_ */
+#endif	/* !__TERM_H_ */
